# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Ovh driver
"""

from libcloud.utils.py3 import httplib
from libcloud.common.ovh import API_ROOT, OvhConnection
from libcloud.compute.base import (
    Node,
    NodeSize,
    NodeImage,
    NodeDriver,
    NodeLocation,
    StorageVolume,
    VolumeSnapshot,
)
from libcloud.compute.types import Provider, StorageVolumeState, VolumeSnapshotState
from libcloud.compute.drivers.openstack import OpenStackKeyPair, OpenStackNodeDriver


class OvhNodeDriver(NodeDriver):
    """
    Libcloud driver for the Ovh API

    For more information on the Ovh API, read the official reference:

        https://api.ovh.com/console/
    """

    type = Provider.OVH
    name = "Ovh"
    website = "https://www.ovh.com/"
    connectionCls = OvhConnection
    features = {"create_node": ["ssh_key"]}
    api_name = "ovh"

    NODE_STATE_MAP = OpenStackNodeDriver.NODE_STATE_MAP
    VOLUME_STATE_MAP = OpenStackNodeDriver.VOLUME_STATE_MAP
    SNAPSHOT_STATE_MAP = OpenStackNodeDriver.SNAPSHOT_STATE_MAP

    def __init__(self, key, secret, ex_project_id, ex_consumer_key=None, region=None):
        """
        Instantiate the driver with the given API credentials.

        :param key: Your application key (required)
        :type key: ``str``

        :param secret: Your application secret (required)
        :type secret: ``str``

        :param ex_project_id: Your project ID
        :type ex_project_id: ``str``

        :param ex_consumer_key: Your consumer key (required)
        :type ex_consumer_key: ``str``

        :param region: The datacenter to connect to (optional)
        :type region: ``str``

        :rtype: ``None``
        """
        self.region = region
        self.project_id = ex_project_id
        self.consumer_key = ex_consumer_key
        NodeDriver.__init__(self, key, secret, ex_consumer_key=ex_consumer_key, region=region)

    def _get_project_action(self, suffix):
        base_url = "{}/cloud/project/{}/".format(API_ROOT, self.project_id)

        return base_url + suffix

    @classmethod
    def list_regions(cls):
        return ["eu", "ca"]

    def list_nodes(self, location=None):
        """
        List all nodes.

        :keyword location: Location (region) used as filter
        :type    location: :class:`NodeLocation`

        :return: List of node objects
        :rtype: ``list`` of :class:`Node`
        """
        action = self._get_project_action("instance")
        data = {}

        if location:
            data["region"] = location.id
        response = self.connection.request(action, data=data)

        return self._to_nodes(response.object)

    def ex_get_node(self, node_id):
        """
        Get a individual node.

        :keyword node_id: Node's ID
        :type    node_id: ``str``

        :return: Created node
        :rtype  : :class:`Node`
        """
        action = self._get_project_action("instance/%s" % node_id)
        response = self.connection.request(action, method="GET")

        return self._to_node(response.object)

    def create_node(self, name, image, size, location, ex_keyname=None):
        """
        Create a new node

        :keyword name: Name of created node
        :type    name: ``str``

        :keyword image: Image used for node
        :type    image: :class:`NodeImage`

        :keyword size: Size (flavor) used for node
        :type    size: :class:`NodeSize`

        :keyword location: Location (region) where to create node
        :type    location: :class:`NodeLocation`

        :keyword ex_keyname: Name of SSH key used
        :type    ex_keyname: ``str``

        :return: Created node
        :rtype : :class:`Node`
        """
        action = self._get_project_action("instance")
        data = {
            "name": name,
            "imageId": image.id,
            "flavorId": size.id,
            "region": location.id,
        }

        if ex_keyname:
            key_id = self.get_key_pair(ex_keyname, location).extra["id"]
            data["sshKeyId"] = key_id
        response = self.connection.request(action, data=data, method="POST")

        return self._to_node(response.object)

    def destroy_node(self, node):
        action = self._get_project_action("instance/%s" % node.id)
        self.connection.request(action, method="DELETE")

        return True

    def list_sizes(self, location=None):
        action = self._get_project_action("flavor")
        params = {}

        if location:
            params["region"] = location.id
        response = self.connection.request(action, params=params)

        return self._to_sizes(response.object)

    def ex_get_size(self, size_id):
        """
        Get an individual size (flavor).

        :keyword size_id: Size's ID
        :type    size_id: ``str``

        :return: Size
        :rtype: :class:`NodeSize`
        """
        action = self._get_project_action("flavor/%s" % size_id)
        response = self.connection.request(action)

        return self._to_size(response.object)

    def list_images(self, location=None, ex_size=None):
        """
        List available images

        :keyword location: Location (region) used as filter
        :type    location: :class:`NodeLocation`

        :keyword ex_size: Exclude images which are incompatible with given size
        :type    ex_size: :class:`NodeImage`

        :return: List of images
        :rtype  : ``list`` of :class:`NodeImage`
        """
        action = self._get_project_action("image")
        params = {}

        if location:
            params["region"] = location.id

        if ex_size:
            params["flavorId"] = ex_size.id
        response = self.connection.request(action, params=params)

        return self._to_images(response.object)

    def get_image(self, image_id):
        action = self._get_project_action("image/%s" % image_id)
        response = self.connection.request(action)

        return self._to_image(response.object)

    def list_locations(self):
        action = self._get_project_action("region")
        data = self.connection.request(action)

        return self._to_locations(data.object)

    def list_key_pairs(self, ex_location=None):
        """
        List available SSH public keys.

        :keyword ex_location: Location (region) used as filter
        :type    ex_location: :class:`NodeLocation`

        :return: Public keys
        :rtype: ``list``of :class:`KeyPair`
        """
        action = self._get_project_action("sshkey")
        params = {}

        if ex_location:
            params["region"] = ex_location.id
        response = self.connection.request(action, params=params)

        return self._to_key_pairs(response.object)

    def get_key_pair(self, name, ex_location=None):
        """
        Get an individual SSH public key by its name and location.

        :param name: Name of the key pair to retrieve.
        :type name: ``str``

        :keyword ex_location: Key's region
        :type ex_location: :class:`NodeLocation`

        :return: Public key
        :rtype: :class:`KeyPair`
        """
        # Keys are indexed with ID
        keys = [key for key in self.list_key_pairs(ex_location) if key.name == name]

        if not keys:
            raise Exception("No key named '%s'" % name)

        return keys[0]

    def import_key_pair_from_string(self, name, key_material, ex_location):
        """
        Import a new public key from string.

        :param name: Key pair name.
        :type name: ``str``

        :param key_material: Public key material.
        :type key_material: ``str``

        :param ex_location: Location where to store the key
        :type ex_location: :class:`NodeLocation`

        :return: Imported key pair object.
        :rtype: :class:`KeyPair`
        """
        action = self._get_project_action("sshkey")
        data = {"name": name, "publicKey": key_material, "region": ex_location.id}
        response = self.connection.request(action, data=data, method="POST")

        return self._to_key_pair(response.object)

    def delete_key_pair(self, key_pair):
        action = self._get_project_action("sshkey/%s" % key_pair.extra["id"])
        params = {"keyId": key_pair.extra["id"]}
        self.connection.request(action, params=params, method="DELETE")

        return True

    def create_volume(
        self,
        size,
        name,
        location,
        snapshot=None,
        ex_volume_type="classic",
        ex_description=None,
    ):
        """
        Create a volume.

        :param size: Size of volume to create (in GB).
        :type size: ``int``

        :param name: Name of volume to create
        :type name: ``str``

        :param location: Location to create the volume in
        :type location: :class:`NodeLocation` or ``None``

        :param snapshot:  Snapshot from which to create the new
                          volume.  (optional)
        :type snapshot: :class:`.VolumeSnapshot`

        :keyword ex_volume_type: ``'classic'`` or ``'high-speed'``
        :type ex_volume_type: ``str``

        :keyword ex_description: Optional description of volume
        :type ex_description: str

        :return:  Storage Volume object
        :rtype:   :class:`StorageVolume`
        """
        action = self._get_project_action("volume")
        data = {
            "name": name,
            "region": location.id,
            "size": size,
            "type": ex_volume_type,
        }

        if ex_description:
            data["description"] = ex_description
        response = self.connection.request(action, data=data, method="POST")

        return self._to_volume(response.object)

    def destroy_volume(self, volume):
        action = self._get_project_action("volume/%s" % volume.id)
        self.connection.request(action, method="DELETE")

        return True

    def list_volumes(self, ex_location=None):
        """
        Return a list of volumes.

        :keyword ex_location: Location used to filter
        :type ex_location: :class:`NodeLocation` or ``None``

        :return: A list of volume objects.
        :rtype: ``list`` of :class:`StorageVolume`
        """
        action = self._get_project_action("volume")
        data = {}

        if ex_location:
            data["region"] = ex_location.id
        response = self.connection.request(action, data=data)

        return self._to_volumes(response.object)

    def ex_get_volume(self, volume_id):
        """
        Return a Volume object based on a volume ID.

        :param  volume_id: The ID of the volume
        :type   volume_id: ``int``

        :return:  A StorageVolume object for the volume
        :rtype:   :class:`StorageVolume`
        """
        action = self._get_project_action("volume/%s" % volume_id)
        response = self.connection.request(action)

        return self._to_volume(response.object)

    def attach_volume(self, node, volume, device=None):
        """
        Attach a volume to a node.

        :param node: Node where to attach volume
        :type node: :class:`Node`

        :param volume: The ID of the volume
        :type volume: :class:`StorageVolume`

        :param device: Unused parameter

        :return: True or False representing operation successful
        :rtype:   ``bool``
        """
        action = self._get_project_action("volume/%s/attach" % volume.id)
        data = {"instanceId": node.id, "volumeId": volume.id}
        self.connection.request(action, data=data, method="POST")

        return True

    def detach_volume(self, volume, ex_node=None):
        """
        Detach a volume to a node.

        :param volume: The ID of the volume
        :type volume: :class:`StorageVolume`

        :param ex_node: Node to detach from (optional if volume is attached
                        to only one node)
        :type ex_node: :class:`Node`

        :return: True or False representing operation successful
        :rtype:   ``bool``

        :raises: Exception: If ``ex_node`` is not provided and more than one
                            node is attached to the volume
        """
        action = self._get_project_action("volume/%s/detach" % volume.id)

        if ex_node is None:
            if len(volume.extra["attachedTo"]) != 1:
                err_msg = (
                    "Volume '%s' has more or less than one attached" "nodes, you must specify one."
                )
                raise Exception(err_msg)
            ex_node = self.ex_get_node(volume.extra["attachedTo"][0])
        data = {"instanceId": ex_node.id}
        self.connection.request(action, data=data, method="POST")

        return True

    def ex_list_snapshots(self, location=None):
        """
        List all snapshots.

        :keyword location: Location used to filter
        :type location: :class:`NodeLocation` or ``None``

        :rtype: ``list`` of :class:`VolumeSnapshot`
        """
        action = self._get_project_action("volume/snapshot")
        params = {}

        if location:
            params["region"] = location.id
        response = self.connection.request(action, params=params)

        return self._to_snapshots(response.object)

    def ex_get_volume_snapshot(self, snapshot_id):
        """
        Returns a single volume snapshot.

        :param snapshot_id: Node to run the task on.
        :type snapshot_id: ``str``

        :rtype :class:`.VolumeSnapshot`:
        :return: Volume snapshot.
        """
        action = self._get_project_action("volume/snapshot/%s" % snapshot_id)
        response = self.connection.request(action)

        return self._to_snapshot(response.object)

    def list_volume_snapshots(self, volume):
        action = self._get_project_action("volume/snapshot")
        params = {"region": volume.extra["region"]}
        response = self.connection.request(action, params=params)
        snapshots = self._to_snapshots(response.object)

        return [snap for snap in snapshots if snap.extra["volume_id"] == volume.id]

    def create_volume_snapshot(self, volume, name=None, ex_description=None):
        """
        Create snapshot from volume

        :param volume: Instance of `StorageVolume`
        :type  volume: `StorageVolume`

        :param name: Name of snapshot (optional)
        :type  name: `str` | `NoneType`

        :param ex_description: Description of the snapshot (optional)
        :type  ex_description: `str` | `NoneType`

        :rtype: :class:`VolumeSnapshot`
        """
        action = self._get_project_action("volume/%s/snapshot/" % volume.id)
        data = {}

        if name:
            data["name"] = name

        if ex_description:
            data["description"] = ex_description
        response = self.connection.request(action, data=data, method="POST")

        return self._to_snapshot(response.object)

    def destroy_volume_snapshot(self, snapshot):
        action = self._get_project_action("volume/snapshot/%s" % snapshot.id)
        response = self.connection.request(action, method="DELETE")

        return response.status == httplib.OK

    def ex_get_pricing(self, size_id, subsidiary="US"):
        action = "%s/cloud/subsidiaryPrice" % (API_ROOT)
        params = {"flavorId": size_id, "ovhSubsidiary": subsidiary}
        pricing = self.connection.request(action, params=params).object["instances"][0]

        return {
            "hourly": pricing["price"]["value"],
            "monthly": pricing["monthlyPrice"]["value"],
        }

    def _to_volume(self, obj):
        extra = obj.copy()
        extra.pop("id")
        extra.pop("name")
        extra.pop("size")
        state = self.VOLUME_STATE_MAP.get(obj.pop("status", None), StorageVolumeState.UNKNOWN)

        return StorageVolume(
            id=obj["id"],
            name=obj["name"],
            size=obj["size"],
            state=state,
            extra=extra,
            driver=self,
        )

    def _to_volumes(self, objs):
        return [self._to_volume(obj) for obj in objs]

    def _to_location(self, obj):
        location = self.connectionCls.LOCATIONS[obj]

        return NodeLocation(driver=self, **location)

    def _to_locations(self, objs):
        return [self._to_location(obj) for obj in objs]

    def _to_node(self, obj):
        extra = obj.copy()

        if "ipAddresses" in extra:
            public_ips = [ip["ip"] for ip in extra["ipAddresses"]]
        else:
            public_ips = []
        del extra["id"]
        del extra["name"]

        return Node(
            id=obj["id"],
            name=obj["name"],
            state=self.NODE_STATE_MAP[obj["status"]],
            public_ips=public_ips,
            private_ips=[],
            driver=self,
            extra=extra,
        )

    def _to_nodes(self, objs):
        return [self._to_node(obj) for obj in objs]

    def _to_size(self, obj):
        extra = {"vcpus": obj["vcpus"], "type": obj["type"], "region": obj["region"]}

        return NodeSize(
            id=obj["id"],
            name=obj["name"],
            ram=obj["ram"],
            disk=obj["disk"],
            bandwidth=obj["outboundBandwidth"],
            price=None,
            driver=self,
            extra=extra,
        )

    def _to_sizes(self, objs):
        return [self._to_size(obj) for obj in objs]

    def _to_image(self, obj):
        extra = {"region": obj["region"], "visibility": obj["visibility"]}

        return NodeImage(id=obj["id"], name=obj["name"], driver=self, extra=extra)

    def _to_images(self, objs):
        return [self._to_image(obj) for obj in objs]

    def _to_key_pair(self, obj):
        extra = {"regions": obj["regions"], "id": obj["id"]}

        return OpenStackKeyPair(
            name=obj["name"],
            public_key=obj["publicKey"],
            driver=self,
            fingerprint=None,
            extra=extra,
        )

    def _to_key_pairs(self, objs):
        return [self._to_key_pair(obj) for obj in objs]

    def _to_snapshot(self, obj):
        extra = {
            "volume_id": obj["volumeId"],
            "region": obj["region"],
            "description": obj["description"],
            "status": obj["status"],
        }
        state = self.SNAPSHOT_STATE_MAP.get(obj["status"], VolumeSnapshotState.UNKNOWN)
        snapshot = VolumeSnapshot(
            id=obj["id"],
            driver=self,
            size=obj["size"],
            extra=extra,
            created=obj["creationDate"],
            state=state,
            name=obj["name"],
        )

        return snapshot

    def _to_snapshots(self, objs):
        return [self._to_snapshot(obj) for obj in objs]

    def _ex_connection_class_kwargs(self):
        return {"ex_consumer_key": self.consumer_key, "region": self.region}



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.compute.base import Node, NodeSize, NodeImage
from libcloud.compute.providers import Provider
from libcloud.compute.drivers.cloudstack import CloudStackNodeDriver


class KTUCloudNodeDriver(CloudStackNodeDriver):
    """Driver for KTUCloud Compute platform."""

    EMPTY_DISKOFFERINGID = "0"
    type = Provider.KTUCLOUD
    name = "KTUCloud"
    website = "https://ucloudbiz.olleh.com/"

    def list_images(self, location=None):
        args = {"templatefilter": "executable"}
        if location is not None:
            args["zoneid"] = location.id

        imgs = self._sync_request(command="listAvailableProductTypes", method="GET")
        images = []

        for img in imgs["producttypes"]:
            images.append(
                NodeImage(
                    img["serviceofferingid"],
                    img["serviceofferingdesc"],
                    self,
                    {
                        "hypervisor": "",
                        "format": "",
                        "os": img["templatedesc"],
                        "templateid": img["templateid"],
                        "zoneid": img["zoneid"],
                    },
                )
            )

        return images

    def list_sizes(self, location=None):
        szs = self._sync_request("listAvailableProductTypes")
        sizes = []
        for sz in szs["producttypes"]:
            diskofferingid = sz.get("diskofferingid", self.EMPTY_DISKOFFERINGID)
            sizes.append(NodeSize(diskofferingid, sz["diskofferingdesc"], 0, 0, 0, 0, self))
        return sizes

    def create_node(self, name, size, image, location=None, ex_usageplantype="hourly"):
        params = {
            "displayname": name,
            "serviceofferingid": image.id,
            "templateid": str(image.extra["templateid"]),
            "zoneid": str(image.extra["zoneid"]),
        }

        if ex_usageplantype is None:
            params["usageplantype"] = "hourly"
        else:
            params["usageplantype"] = ex_usageplantype

        if size.id != self.EMPTY_DISKOFFERINGID:
            params["diskofferingid"] = size.id

        result = self._async_request(command="deployVirtualMachine", params=params, method="GET")

        node = result["virtualmachine"]

        return Node(
            id=node["id"],
            name=node["displayname"],
            state=self.NODE_STATE_MAP[node["state"]],
            public_ips=[],
            private_ips=[],
            driver=self,
            extra={
                "zoneid": image.extra["zoneid"],
                "ip_addresses": [],
                "forwarding_rules": [],
            },
        )



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Drivers for CloudSigma API v1.0 and v2.0.
"""

import re
import copy
import time
import base64
import hashlib

from libcloud.utils.py3 import b, httplib
from libcloud.utils.misc import dict2str, str2list, str2dicts, get_secure_random_string
from libcloud.common.base import Response, JsonResponse, ConnectionUserAndKey
from libcloud.common.types import ProviderError, InvalidCredsError
from libcloud.compute.base import Node, KeyPair, NodeSize, NodeImage, NodeDriver, is_private_subnet
from libcloud.compute.types import Provider, NodeState
from libcloud.utils.iso8601 import parse_date
from libcloud.common.cloudsigma import (
    SPECS_TO_SIZE,
    DEFAULT_REGION,
    INSTANCE_TYPES,
    MAX_VIRTIO_UNITS,
    API_ENDPOINTS_1_0,
    API_ENDPOINTS_2_0,
    DEFAULT_API_VERSION,
    MAX_VIRTIO_CONTROLLERS,
)

try:
    import simplejson as json
except Exception:
    import json


__all__ = [
    "CloudSigmaNodeDriver",
    "CloudSigma_1_0_NodeDriver",
    "CloudSigma_2_0_NodeDriver",
    "CloudSigmaError",
    "CloudSigmaNodeSize",
    "CloudSigmaDrive",
    "CloudSigmaTag",
    "CloudSigmaSubscription",
    "CloudSigmaFirewallPolicy",
    "CloudSigmaFirewallPolicyRule",
]


class CloudSigmaNodeDriver(NodeDriver):
    name = "CloudSigma"
    website = "http://www.cloudsigma.com/"

    def __new__(
        cls,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        api_version=DEFAULT_API_VERSION,
        **kwargs,
    ):
        if cls is CloudSigmaNodeDriver:
            if api_version == "1.0":
                cls = CloudSigma_1_0_NodeDriver
            elif api_version == "2.0":
                cls = CloudSigma_2_0_NodeDriver
            else:
                raise NotImplementedError("Unsupported API version: %s" % (api_version))

        return super().__new__(cls)


class CloudSigmaException(Exception):
    def __str__(self):
        # pylint: disable=unsubscriptable-object

        return self.args[0]

    def __repr__(self):
        # pylint: disable=unsubscriptable-object

        return "<CloudSigmaException '%s'>" % (self.args[0])


class CloudSigmaInsufficientFundsException(Exception):
    def __repr__(self):
        # pylint: disable=unsubscriptable-object

        return "<CloudSigmaInsufficientFundsException '%s'>" % (self.args[0])


class CloudSigmaNodeSize(NodeSize):
    def __init__(self, id, name, cpu, ram, disk, bandwidth, price, driver, extra=None):
        self.id = id
        self.name = name
        self.cpu = cpu
        self.ram = ram
        self.disk = disk
        self.bandwidth = bandwidth
        self.price = price
        self.driver = driver
        self.extra = extra or {}

    def __repr__(self):
        return (
            "<NodeSize: id=%s, name=%s, cpu=%s, ram=%s disk=%s "
            "bandwidth=%s price=%s driver=%s ...>"
        ) % (
            self.id,
            self.name,
            self.cpu,
            self.ram,
            self.disk,
            self.bandwidth,
            self.price,
            self.driver.name,
        )


class CloudSigma_1_0_Response(Response):
    def success(self):
        if self.status == httplib.UNAUTHORIZED:
            raise InvalidCredsError()

        return 200 <= self.status <= 299

    def parse_body(self):
        if not self.body:
            return self.body

        return str2dicts(self.body)

    def parse_error(self):
        return "Error: %s" % (self.body.replace("errors:", "").strip())


class CloudSigma_1_0_Connection(ConnectionUserAndKey):
    host = API_ENDPOINTS_1_0[DEFAULT_REGION]["host"]
    responseCls = CloudSigma_1_0_Response

    def add_default_headers(self, headers):
        headers["Accept"] = "application/json"
        headers["Content-Type"] = "application/json"

        headers["Authorization"] = "Basic %s" % (
            base64.b64encode(b("{}:{}".format(self.user_id, self.key))).decode("utf-8")
        )

        return headers


class CloudSigma_1_0_NodeDriver(CloudSigmaNodeDriver):
    type = Provider.CLOUDSIGMA
    name = "CloudSigma (API v1.0)"
    website = "http://www.cloudsigma.com/"
    connectionCls = CloudSigma_1_0_Connection

    IMAGING_TIMEOUT = 20 * 60  # Default timeout (in seconds) for the drive
    # imaging process

    NODE_STATE_MAP = {
        "active": NodeState.RUNNING,
        "stopped": NodeState.TERMINATED,
        "dead": NodeState.TERMINATED,
        "dumped": NodeState.TERMINATED,
    }

    def __init__(
        self,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        region=DEFAULT_REGION,
        **kwargs,
    ):
        if region not in API_ENDPOINTS_1_0:
            raise ValueError("Invalid region: %s" % (region))

        self._host_argument_set = host is not None
        self.api_name = "cloudsigma_%s" % (region)
        super().__init__(
            key=key,
            secret=secret,
            secure=secure,
            host=host,
            port=port,
            region=region,
            **kwargs,
        )

    def reboot_node(self, node):
        """
        Reboot a node.

        Because Cloudsigma API does not provide native reboot call,
        it's emulated using stop and start.

        @inherits: :class:`NodeDriver.reboot_node`
        """
        node = self._get_node(node.id)
        state = node.state

        if state == NodeState.RUNNING:
            stopped = self.ex_stop_node(node)
        else:
            stopped = True

        if not stopped:
            raise CloudSigmaException("Could not stop node with id %s" % (node.id))

        success = self.ex_start_node(node)

        return success

    def destroy_node(self, node):
        """
        Destroy a node (all the drives associated with it are NOT destroyed).

        If a node is still running, it's stopped before it's destroyed.

        @inherits: :class:`NodeDriver.destroy_node`
        """
        node = self._get_node(node.id)
        state = node.state

        # Node cannot be destroyed while running so it must be stopped first

        if state == NodeState.RUNNING:
            stopped = self.ex_stop_node(node)
        else:
            stopped = True

        if not stopped:
            raise CloudSigmaException("Could not stop node with id %s" % (node.id))

        response = self.connection.request(action="/servers/%s/destroy" % (node.id), method="POST")

        return response.status == 204

    def list_images(self, location=None):
        """
        Return a list of available standard images (this call might take up
        to 15 seconds to return).

        @inherits: :class:`NodeDriver.list_images`
        """
        response = self.connection.request(action="/drives/standard/info").object

        images = []

        for value in response:
            if value.get("type"):
                if value["type"] == "disk":
                    image = NodeImage(
                        id=value["drive"],
                        name=value["name"],
                        driver=self.connection.driver,
                        extra={"size": value["size"]},
                    )
                    images.append(image)

        return images

    def list_sizes(self, location=None):
        sizes = []

        for value in INSTANCE_TYPES:
            key = value["id"]
            size = CloudSigmaNodeSize(
                id=value["id"],
                name=value["name"],
                cpu=value["cpu"],
                ram=value["memory"],
                disk=value["disk"],
                bandwidth=value["bandwidth"],
                price=self._get_size_price(size_id=key),
                driver=self.connection.driver,
            )
            sizes.append(size)

        return sizes

    def list_nodes(self):
        response = self.connection.request(action="/servers/info").object

        nodes = []

        for data in response:
            node = self._to_node(data)

            if node:
                nodes.append(node)

        return nodes

    def create_node(
        self,
        name,
        size,
        image,
        smp="auto",
        nic_model="e1000",
        vnc_password=None,
        drive_type="hdd",
    ):
        """
        Creates a CloudSigma instance

        @inherits: :class:`NodeDriver.create_node`

        :keyword    name: String with a name for this new node (required)
        :type       name: ``str``

        :keyword    smp: Number of virtual processors or None to calculate
                         based on the cpu speed.
        :type       smp: ``int``

        :keyword    nic_model: e1000, rtl8139 or virtio (is not specified,
                               e1000 is used)
        :type       nic_model: ``str``

        :keyword    vnc_password: If not set, VNC access is disabled.
        :type       vnc_password: ``bool``

        :keyword    drive_type: Drive type (ssd|hdd). Defaults to hdd.
        :type       drive_type: ``str``
        """

        if nic_model not in ["e1000", "rtl8139", "virtio"]:
            raise CloudSigmaException("Invalid NIC model specified")

        if drive_type not in ["hdd", "ssd"]:
            raise CloudSigmaException(
                'Invalid drive type "%s". Valid types' " are: hdd, ssd" % (drive_type)
            )

        drive_data = {}
        drive_data.update({"name": name, "size": "%sG" % (size.disk), "driveType": drive_type})

        response = self.connection.request(
            action="/drives/%s/clone" % image.id,
            data=dict2str(drive_data),
            method="POST",
        ).object

        if not response:
            raise CloudSigmaException("Drive creation failed")

        drive_uuid = response[0]["drive"]

        response = self.connection.request(action="/drives/%s/info" % (drive_uuid)).object
        imaging_start = time.time()

        while "imaging" in response[0]:
            response = self.connection.request(action="/drives/%s/info" % (drive_uuid)).object
            elapsed_time = time.time() - imaging_start
            timed_out = elapsed_time >= self.IMAGING_TIMEOUT

            if "imaging" in response[0] and timed_out:
                raise CloudSigmaException("Drive imaging timed out")
            time.sleep(1)

        node_data = {}
        node_data.update(
            {
                "name": name,
                "cpu": size.cpu,
                "mem": size.ram,
                "ide:0:0": drive_uuid,
                "boot": "ide:0:0",
                "smp": smp,
            }
        )
        node_data.update({"nic:0:model": nic_model, "nic:0:dhcp": "auto"})

        if vnc_password:
            node_data.update({"vnc:ip": "auto", "vnc:password": vnc_password})

        response = self.connection.request(
            action="/servers/create", data=dict2str(node_data), method="POST"
        ).object

        if not isinstance(response, list):
            response = [response]

        node = self._to_node(response[0])

        if node is None:
            # Insufficient funds, destroy created drive
            self.ex_drive_destroy(drive_uuid)
            raise CloudSigmaInsufficientFundsException("Insufficient funds, node creation failed")

        # Start the node after it has been created
        started = self.ex_start_node(node)

        if started:
            node.state = NodeState.RUNNING

        return node

    def ex_destroy_node_and_drives(self, node):
        """
        Destroy a node and all the drives associated with it.

        :param      node: Node which should be used
        :type       node: :class:`libcloud.compute.base.Node`

        :rtype: ``bool``
        """
        node = self._get_node_info(node)

        drive_uuids = []

        for key, value in node.items():
            if (
                key.startswith("ide:") or key.startswith("scsi") or key.startswith("block")
            ) and not (
                key.endswith(":bytes") or key.endswith(":requests") or key.endswith("media")
            ):
                drive_uuids.append(value)

        node_destroyed = self.destroy_node(self._to_node(node))

        if not node_destroyed:
            return False

        for drive_uuid in drive_uuids:
            self.ex_drive_destroy(drive_uuid)

        return True

    def ex_static_ip_list(self):
        """
        Return a list of available static IP addresses.

        :rtype: ``list`` of ``str``
        """
        response = self.connection.request(action="/resources/ip/list", method="GET")

        if response.status != 200:
            raise CloudSigmaException("Could not retrieve IP list")

        ips = str2list(response.body)

        return ips

    def ex_drives_list(self):
        """
        Return a list of all the available drives.

        :rtype: ``list`` of ``dict``
        """
        response = self.connection.request(action="/drives/info", method="GET")

        result = str2dicts(response.body)

        return result

    def ex_static_ip_create(self):
        """
        Create a new static IP address.p

        :rtype: ``list`` of ``dict``
        """
        response = self.connection.request(action="/resources/ip/create", method="GET")

        result = str2dicts(response.body)

        return result

    def ex_static_ip_destroy(self, ip_address):
        """
        Destroy a static IP address.

        :param      ip_address: IP address which should be used
        :type       ip_address: ``str``

        :rtype: ``bool``
        """
        response = self.connection.request(
            action="/resources/ip/%s/destroy" % (ip_address), method="GET"
        )

        return response.status == 204

    def ex_drive_destroy(self, drive_uuid):
        """
        Destroy a drive with a specified uuid.
        If the drive is currently mounted an exception is thrown.

        :param      drive_uuid: Drive uuid which should be used
        :type       drive_uuid: ``str``

        :rtype: ``bool``
        """
        response = self.connection.request(
            action="/drives/%s/destroy" % (drive_uuid), method="POST"
        )

        return response.status == 204

    def ex_set_node_configuration(self, node, **kwargs):
        """
        Update a node configuration.
        Changing most of the parameters requires node to be stopped.

        :param      node: Node which should be used
        :type       node: :class:`libcloud.compute.base.Node`

        :param      kwargs: keyword arguments
        :type       kwargs: ``dict``

        :rtype: ``bool``
        """
        valid_keys = (
            "^name$",
            "^parent$",
            "^cpu$",
            "^smp$",
            "^mem$",
            "^boot$",
            "^nic:0:model$",
            "^nic:0:dhcp",
            "^nic:1:model$",
            "^nic:1:vlan$",
            "^nic:1:mac$",
            "^vnc:ip$",
            "^vnc:password$",
            "^vnc:tls",
            "^ide:[0-1]:[0-1](:media)?$",
            "^scsi:0:[0-7](:media)?$",
            "^block:[0-7](:media)?$",
        )

        invalid_keys = []
        keys = list(kwargs.keys())

        for key in keys:
            matches = False

            for regex in valid_keys:
                if re.match(regex, key):
                    matches = True

                    break

            if not matches:
                invalid_keys.append(key)

        if invalid_keys:
            raise CloudSigmaException(
                "Invalid configuration key specified: %s" % (",".join(invalid_keys))
            )

        response = self.connection.request(
            action="/servers/%s/set" % (node.id), data=dict2str(kwargs), method="POST"
        )

        return response.status == 200 and response.body != ""

    def ex_start_node(self, node):
        """
        Start a node.

        :param      node: Node which should be used
        :type       node: :class:`libcloud.compute.base.Node`

        :rtype: ``bool``
        """
        response = self.connection.request(action="/servers/%s/start" % (node.id), method="POST")

        return response.status == 200

    def ex_stop_node(self, node):
        # NOTE: This method is here for backward compatibility reasons after
        # this method was promoted to be part of the standard compute API in
        # Libcloud v2.7.0

        return self.stop_node(node=node)

    def stop_node(self, node):
        """
        Stop (shutdown) a node.

        :param      node: Node which should be used
        :type       node: :class:`libcloud.compute.base.Node`

        :rtype: ``bool``
        """
        response = self.connection.request(action="/servers/%s/stop" % (node.id), method="POST")

        return response.status == 204

    def ex_shutdown_node(self, node):
        """
        Stop (shutdown) a node.

        @inherits: :class:`CloudSigmaBaseNodeDriver.ex_stop_node`
        """

        return self.ex_stop_node(node)

    def ex_destroy_drive(self, drive_uuid):
        """
        Destroy a drive.

        :param      drive_uuid: Drive uuid which should be used
        :type       drive_uuid: ``str``

        :rtype: ``bool``
        """
        response = self.connection.request(
            action="/drives/%s/destroy" % (drive_uuid), method="POST"
        )

        return response.status == 204

    def _ex_connection_class_kwargs(self):
        """
        Return the host value based on the user supplied region.
        """
        kwargs = {}

        if not self._host_argument_set:
            kwargs["host"] = API_ENDPOINTS_1_0[self.region]["host"]

        return kwargs

    def _to_node(self, data):
        if data:
            try:
                state = self.NODE_STATE_MAP[data["status"]]
            except KeyError:
                state = NodeState.UNKNOWN

            if "server" not in data:
                # Response does not contain server UUID if the server
                # creation failed because of insufficient funds.

                return None

            public_ips = []

            if "nic:0:dhcp" in data:
                if isinstance(data["nic:0:dhcp"], list):
                    public_ips = data["nic:0:dhcp"]
                else:
                    public_ips = [data["nic:0:dhcp"]]

            extra = {}
            extra_keys = [
                ("cpu", "int"),
                ("smp", "auto"),
                ("mem", "int"),
                ("status", "str"),
            ]

            for key, value_type in extra_keys:
                if key in data:
                    value = data[key]

                    if value_type == "int":
                        value = int(value)
                    elif value_type == "auto":
                        try:
                            value = int(value)
                        except ValueError:
                            pass

                    extra.update({key: value})

            if "vnc:ip" in data and "vnc:password" in data:
                extra.update({"vnc_ip": data["vnc:ip"], "vnc_password": data["vnc:password"]})

            node = Node(
                id=data["server"],
                name=data["name"],
                state=state,
                public_ips=public_ips,
                private_ips=None,
                driver=self.connection.driver,
                extra=extra,
            )

            return node

        return None

    def _get_node(self, node_id):
        nodes = self.list_nodes()
        node = [node for node in nodes if node.id == node.id]

        if not node:
            raise CloudSigmaException("Node with id %s does not exist" % (node_id))

        return node[0]

    def _get_node_info(self, node):
        response = self.connection.request(action="/servers/%s/info" % (node.id))

        result = str2dicts(response.body)

        return result[0]


class CloudSigmaZrhConnection(CloudSigma_1_0_Connection):
    """
    Connection class for the CloudSigma driver for the Zurich end-point
    """

    host = API_ENDPOINTS_1_0["zrh"]["host"]


class CloudSigmaZrhNodeDriver(CloudSigma_1_0_NodeDriver):
    """
    CloudSigma node driver for the Zurich end-point
    """

    connectionCls = CloudSigmaZrhConnection
    api_name = "cloudsigma_zrh"


class CloudSigmaLvsConnection(CloudSigma_1_0_Connection):
    """
    Connection class for the CloudSigma driver for the Las Vegas end-point
    """

    host = API_ENDPOINTS_1_0["lvs"]["host"]


class CloudSigmaLvsNodeDriver(CloudSigma_1_0_NodeDriver):
    """
    CloudSigma node driver for the Las Vegas end-point
    """

    connectionCls = CloudSigmaLvsConnection
    api_name = "cloudsigma_lvs"


class CloudSigmaError(ProviderError):
    """
    Represents CloudSigma API error.
    """

    def __init__(self, http_code, error_type, error_msg, error_point, driver):
        """
        :param http_code: HTTP status code.
        :type http_code: ``int``

        :param error_type: Type of error (validation / notexist / backend /
                           permissions  database / concurrency / billing /
                           payment)
        :type error_type: ``str``

        :param error_msg: A description of the error that occurred.
        :type error_msg: ``str``

        :param error_point: Point at which the error occurred. Can be None.
        :type error_point: ``str`` or ``None``
        """
        super().__init__(http_code=http_code, value=error_msg, driver=driver)
        self.error_type = error_type
        self.error_msg = error_msg
        self.error_point = error_point


class CloudSigmaSubscription:
    """
    Represents CloudSigma subscription.
    """

    def __init__(
        self,
        id,
        resource,
        amount,
        period,
        status,
        price,
        start_time,
        end_time,
        auto_renew,
        subscribed_object=None,
    ):
        """
        :param id: Subscription ID.
        :type id: ``str``

        :param resource: Resource (e.g vlan, ip, etc.).
        :type resource: ``str``

        :param period: Subscription period.
        :type period: ``str``

        :param status: Subscription status (active / inactive).
        :type status: ``str``

        :param price: Subscription price.
        :type price: ``str``

        :param start_time: Start time for this subscription.
        :type start_time: ``datetime.datetime``

        :param end_time: End time for this subscription.
        :type end_time: ``datetime.datetime``

        :param auto_renew: True if the subscription is auto renewed.
        :type auto_renew: ``bool``

        :param subscribed_object: Optional UUID of the subscribed object.
        :type subscribed_object: ``str``
        """
        self.id = id
        self.resource = resource
        self.amount = amount
        self.period = period
        self.status = status
        self.price = price
        self.start_time = start_time
        self.end_time = end_time
        self.auto_renew = auto_renew
        self.subscribed_object = subscribed_object

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return (
            "<CloudSigmaSubscription id=%s, resource=%s, amount=%s, "
            "period=%s, object_uuid=%s>"
            % (self.id, self.resource, self.amount, self.period, self.subscribed_object)
        )


class CloudSigmaTag:
    """
    Represents a CloudSigma tag object.
    """

    def __init__(self, id, name, resources=None):
        """
        :param id: Tag ID.
        :type id: ``str``

        :param name: Tag name.
        :type name: ``str``

        :param resource: IDs of resources which are associated with this tag.
        :type resources: ``list`` of ``str``
        """
        self.id = id
        self.name = name
        self.resources = resources if resources else []

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<CloudSigmaTag id={}, name={}, resources={}>".format(
            self.id,
            self.name,
            repr(self.resources),
        )


class CloudSigmaDrive(NodeImage):
    """
    Represents a CloudSigma drive.
    """

    def __init__(self, id, name, size, media, status, driver, extra=None):
        """
        :param id: Drive ID.
        :type id: ``str``

        :param name: Drive name.
        :type name: ``str``

        :param size: Drive size (in GBs).
        :type size: ``float``

        :param media: Drive media (cdrom / disk).
        :type media: ``str``

        :param status: Drive status (unmounted / mounted).
        :type status: ``str``
        """
        super().__init__(id=id, name=name, driver=driver, extra=extra)
        self.size = size
        self.media = media
        self.status = status

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return ("<CloudSigmaSize id=%s, name=%s size=%s, media=%s, " "status=%s>") % (
            self.id,
            self.name,
            self.size,
            self.media,
            self.status,
        )


class CloudSigmaFirewallPolicy:
    """
    Represents a CloudSigma firewall policy.
    """

    def __init__(self, id, name, rules):
        """
        :param id: Policy ID.
        :type id: ``str``

        :param name: Policy name.
        :type name: ``str``

        :param rules: Rules associated with this policy.
        :type rules: ``list`` of :class:`.CloudSigmaFirewallPolicyRule` objects
        """
        self.id = id
        self.name = name
        self.rules = rules if rules else []

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return ("<CloudSigmaFirewallPolicy id=%s, name=%s rules=%s>") % (
            self.id,
            self.name,
            repr(self.rules),
        )


class CloudSigmaFirewallPolicyRule:
    """
    Represents a CloudSigma firewall policy rule.
    """

    def __init__(
        self,
        action,
        direction,
        ip_proto=None,
        src_ip=None,
        src_port=None,
        dst_ip=None,
        dst_port=None,
        comment=None,
    ):
        """
        :param action: Action (drop / accept).
        :type action: ``str``

        :param direction: Rule direction (in / out / both)>
        :type direction: ``str``

        :param ip_proto: IP protocol (tcp / udp).
        :type ip_proto: ``str``.

        :param src_ip: Source IP in CIDR notation.
        :type src_ip: ``str``

        :param src_port: Source port or a port range.
        :type src_port: ``str``

        :param dst_ip: Destination IP in CIDR notation.
        :type dst_ip: ``str``

        :param src_port: Destination port or a port range.
        :type src_port: ``str``

        :param comment: Comment associated with the policy.
        :type comment: ``str``
        """
        self.action = action
        self.direction = direction
        self.ip_proto = ip_proto
        self.src_ip = src_ip
        self.src_port = src_port
        self.dst_ip = dst_ip
        self.dst_port = dst_port
        self.comment = comment

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return ("<CloudSigmaFirewallPolicyRule action=%s, direction=%s>") % (
            self.action,
            self.direction,
        )


class CloudSigma_2_0_Response(JsonResponse):
    success_status_codes = [
        httplib.OK,
        httplib.ACCEPTED,
        httplib.NO_CONTENT,
        httplib.CREATED,
    ]

    def success(self):
        return self.status in self.success_status_codes

    def parse_error(self):
        if int(self.status) == httplib.UNAUTHORIZED:
            raise InvalidCredsError("Invalid credentials")

        body = self.parse_body()
        errors = self._parse_errors_from_body(body=body)

        if errors:
            # Throw first error
            raise errors[0]

        return body

    def _parse_errors_from_body(self, body):
        """
        Parse errors from the response body.

        :return: List of error objects.
        :rtype: ``list`` of :class:`.CloudSigmaError` objects
        """
        errors = []

        if not isinstance(body, list):
            return None

        for item in body:
            if "error_type" not in item:
                # Unrecognized error

                continue

            error = CloudSigmaError(
                http_code=self.status,
                error_type=item["error_type"],
                error_msg=item["error_message"],
                error_point=item["error_point"],
                driver=self.connection.driver,
            )
            errors.append(error)

        return errors


class CloudSigma_2_0_Connection(ConnectionUserAndKey):
    host = API_ENDPOINTS_2_0[DEFAULT_REGION]["host"]
    responseCls = CloudSigma_2_0_Response
    api_prefix = "/api/2.0"

    def add_default_headers(self, headers):
        headers["Accept"] = "application/json"
        headers["Content-Type"] = "application/json"

        headers["Authorization"] = "Basic %s" % (
            base64.b64encode(b("{}:{}".format(self.user_id, self.key))).decode("utf-8")
        )

        return headers

    def encode_data(self, data):
        data = json.dumps(data)

        return data

    def request(self, action, params=None, data=None, headers=None, method="GET", raw=False):
        params = params or {}
        action = self.api_prefix + action

        if method == "GET":
            params["limit"] = 0  # we want all the items back

        return super().request(
            action=action,
            params=params,
            data=data,
            headers=headers,
            method=method,
            raw=raw,
        )


class CloudSigma_2_0_NodeDriver(CloudSigmaNodeDriver):
    """
    Driver for CloudSigma API v2.0.
    """

    name = "CloudSigma (API v2.0)"
    api_name = "cloudsigma_zrh"
    website = "http://www.cloudsigma.com/"
    connectionCls = CloudSigma_2_0_Connection

    # Default drive transition timeout in seconds
    DRIVE_TRANSITION_TIMEOUT = 500

    # How long to sleep between different polling periods while waiting for
    # drive transition
    DRIVE_TRANSITION_SLEEP_INTERVAL = 5

    NODE_STATE_MAP = {
        "starting": NodeState.PENDING,
        "stopping": NodeState.PENDING,
        "unavailable": NodeState.ERROR,
        "running": NodeState.RUNNING,
        "stopped": NodeState.STOPPED,
        "paused": NodeState.PAUSED,
    }

    def __init__(
        self,
        key,
        secret,
        secure=True,
        host=None,
        port=None,
        region=DEFAULT_REGION,
        **kwargs,
    ):
        if region not in API_ENDPOINTS_2_0:
            raise ValueError("Invalid region: %s" % (region))

        if not secure:
            # CloudSigma drive uses Basic Auth authentication and we don't want
            # to allow user to accidentally send credentials over the wire in
            # plain-text
            raise ValueError("CloudSigma driver only supports a " "secure connection")

        self._host_argument_set = host is not None
        super().__init__(
            key=key,
            secret=secret,
            secure=secure,
            host=host,
            port=port,
            region=region,
            **kwargs,
        )

    def list_nodes(self, ex_tag=None):
        """
        List available nodes.

        :param ex_tag: If specified, only return servers tagged with the
                       provided tag.
        :type ex_tag: :class:`CloudSigmaTag`
        """

        if ex_tag:
            action = "/tags/%s/servers/detail/" % (ex_tag.id)
        else:
            action = "/servers/detail/"

        response = self.connection.request(action=action, method="GET").object
        nodes = [self._to_node(data=item) for item in response["objects"]]

        return nodes

    def list_sizes(self):
        """
        List available sizes.
        """
        sizes = []

        for value in INSTANCE_TYPES:
            key = value["id"]
            size = CloudSigmaNodeSize(
                id=value["id"],
                name=value["name"],
                cpu=value["cpu"],
                ram=value["memory"],
                disk=value["disk"],
                bandwidth=value["bandwidth"],
                price=self._get_size_price(size_id=key),
                driver=self.connection.driver,
            )
            sizes.append(size)

        return sizes

    def list_images(self):
        """
        Return a list of available pre-installed library drives.

        Note: If you want to list all the available library drives (both
        pre-installed and installation CDs), use :meth:`ex_list_library_drives`
        method.
        """
        response = self.connection.request(action="/libdrives/").object
        images = [self._to_image(data=item) for item in response["objects"]]

        # We filter out non pre-installed library drives by default because
        # they can't be used directly following a default Libcloud server
        # creation flow.
        images = [image for image in images if image.extra["image_type"] == "preinst"]

        return images

    def create_node(
        self,
        name,
        size,
        image,
        ex_metadata=None,
        ex_vnc_password=None,
        ex_avoid=None,
        ex_vlan=None,
        public_keys=None,
    ):
        """
        Create a new server.

        Server creation consists multiple steps depending on the type of the
        image used.

        1. Installation CD:

            1. Create a server and attach installation cd
            2. Start a server

        2. Pre-installed image:

            1. Clone provided library drive so we can use it
            2. Resize cloned drive to the desired size
            3. Create a server and attach cloned drive
            4. Start a server

        :param ex_metadata: Key / value pairs to associate with the
                            created node. (optional)
        :type ex_metadata: ``dict``

        :param ex_vnc_password: Password to use for VNC access. If not
                                provided, random password is generated.
        :type ex_vnc_password: ``str``

        :param ex_avoid: A list of server UUIDs to avoid when starting this
                         node. (optional)
        :type ex_avoid: ``list``

        :param ex_vlan: Optional UUID of a VLAN network to use. If specified,
                        server will have two nics assigned - 1 with a public ip
                        and 1 with the provided VLAN.
        :type ex_vlan: ``str``

        :param public_keys: Optional list of SSH key UUIDs
        :type public_keys: ``list`` of ``str``
        """
        is_installation_cd = self._is_installation_cd(image=image)

        if ex_vnc_password:
            vnc_password = ex_vnc_password
        else:
            # VNC password is not provided, generate a random one.
            vnc_password = get_secure_random_string(size=12)

        drive_name = "%s-drive" % (name)

        drive_size = size.disk

        if not is_installation_cd:
            # 1. Clone library drive so we can use it
            drive = self.ex_clone_drive(drive=image, name=drive_name)

            # Wait for drive clone to finish
            drive = self._wait_for_drive_state_transition(drive=drive, state="unmounted")

            # 2. Resize drive to the desired disk size if the desired disk size
            # is larger than the cloned drive size.

            if drive_size > drive.size:
                drive = self.ex_resize_drive(drive=drive, size=drive_size)

            # Wait for drive resize to finish
            drive = self._wait_for_drive_state_transition(drive=drive, state="unmounted")
        else:
            # No need to clone installation CDs
            drive = image

        # 3. Create server and attach cloned drive
        # ide 0:0
        data = {}
        data["name"] = name
        # cloudsigma uses MHz to measure cpu cores,
        # where 1 cpu core equals 2000MHz
        data["cpu"] = size.cpu * 2000
        data["mem"] = size.ram * 1024 * 1024
        data["vnc_password"] = vnc_password

        if public_keys:
            data["pubkeys"] = public_keys

        if ex_metadata:
            data["meta"] = ex_metadata

        # Assign 1 public interface (DHCP) to the node
        nic = {"boot_order": None, "ip_v4_conf": {"conf": "dhcp"}, "ip_v6_conf": None}

        nics = [nic]

        if ex_vlan:
            # Assign another interface for VLAN
            nic = {
                "boot_order": None,
                "ip_v4_conf": None,
                "ip_v6_conf": None,
                "vlan": ex_vlan,
            }
            nics.append(nic)

        # Need to use IDE for installation CDs

        if is_installation_cd:
            device_type = "ide"
        else:
            device_type = "virtio"

        drive = {
            "boot_order": 1,
            "dev_channel": "0:0",
            "device": device_type,
            "drive": drive.id,
        }

        drives = [drive]

        data["nics"] = nics
        data["drives"] = drives

        action = "/servers/"
        response = self.connection.request(action=action, method="POST", data=data)
        node = self._to_node(response.object["objects"][0])

        # 4. Start server
        self.ex_start_node(node=node, ex_avoid=ex_avoid)

        return node

    def destroy_node(self, node, ex_delete_drives=False):
        """
        Destroy the node and all the associated drives.

        :return: ``True`` on success, ``False`` otherwise.
        :rtype: ``bool``
        """
        action = "/servers/%s/" % (node.id)

        if ex_delete_drives is True:
            params = {"recurse": "all_drives"}
        else:
            params = None
        response = self.connection.request(action=action, method="DELETE", params=params)

        return response.status == httplib.NO_CONTENT

    def reboot_node(self, node):
        """
        Reboot a node.

        Because Cloudsigma API does not provide native reboot call,
        it's emulated using stop and start.

        :param node: Node to reboot.
        :type node: :class:`libcloud.compute.base.Node`
        """
        state = node.state

        if state == NodeState.RUNNING:
            stopped = self.stop_node(node)
        else:
            stopped = True

        if not stopped:
            raise CloudSigmaException("Could not stop node with id %s" % (node.id))

        success = False

        for _ in range(5):
            try:
                success = self.start_node(node)
            except CloudSigmaError:
                time.sleep(1)

                continue
            else:
                break

        return success

    # Server extension methods

    def ex_edit_node(self, node, params):
        """
        Edit a node.

        :param node: Node to edit.
        :type node: :class:`libcloud.compute.base.Node`

        :param params: Node parameters to update.
        :type params: ``dict``

        :return Edited node.
        :rtype: :class:`libcloud.compute.base.Node`
        """
        data = {}

        # name, cpu, mem and vnc_password attributes must always be present so
        # we just copy them from the to-be-edited node
        data["name"] = node.name
        data["cpu"] = node.extra["cpus"]
        data["mem"] = node.extra["memory"]
        data["vnc_password"] = node.extra["vnc_password"]

        nics = copy.deepcopy(node.extra.get("nics", []))

        data["nics"] = nics

        data.update(params)

        action = "/servers/%s/" % (node.id)
        response = self.connection.request(action=action, method="PUT", data=data).object
        node = self._to_node(data=response)

        return node

    def start_node(self, node, ex_avoid=None):
        """
        Start a node.

        :param node: Node to start.
        :type node: :class:`libcloud.compute.base.Node`

        :param ex_avoid: A list of other server uuids to avoid when
                         starting this node. If provided, node will
                         attempt to be started on a different
                         physical infrastructure from other servers
                         specified using this argument. (optional)
        :type ex_avoid: ``list``
        """
        params = {}

        if ex_avoid:
            params["avoid"] = ",".join(ex_avoid)

        path = "/servers/%s/action/" % (node.id)
        response = self._perform_action(path=path, action="start", params=params, method="POST")

        return response.status == httplib.ACCEPTED

    def stop_node(self, node):
        path = "/servers/%s/action/" % (node.id)
        response = self._perform_action(path=path, action="stop", method="POST")

        return response.status == httplib.ACCEPTED

    def ex_start_node(self, node, ex_avoid=None):
        # NOTE: This method is here for backward compatibility reasons after
        # this method was promoted to be part of the standard compute API in
        # Libcloud v2.7.0

        return self.start_node(node=node, ex_avoid=ex_avoid)

    def ex_stop_node(self, node):
        # NOTE: This method is here for backward compatibility reasons after
        # this method was promoted to be part of the standard compute API in
        # Libcloud v2.7.0
        """
        Stop a node.
        """

        return self.stop_node(node=node)

    def ex_clone_node(self, node, name=None, random_vnc_password=None):
        """
        Clone the provided node.

        :param name: Optional name for the cloned node.
        :type name: ``str``
        :param random_vnc_password: If True, a new random VNC password will be
                                    generated for the cloned node. Otherwise
                                    password from the cloned node will be
                                    reused.
        :type random_vnc_password: ``bool``

        :return: Cloned node.
        :rtype: :class:`libcloud.compute.base.Node`
        """
        data = {}

        data["name"] = name
        data["random_vnc_password"] = random_vnc_password

        path = "/servers/%s/action/" % (node.id)
        response = self._perform_action(path=path, action="clone", method="POST", data=data).object
        node = self._to_node(data=response)

        return node

    def ex_get_node(self, node_id, return_json=False):
        action = "/servers/%s/" % (node_id)
        response = self.connection.request(action=action).object

        if return_json is True:
            return response

        return self._to_node(response)

    def ex_open_vnc_tunnel(self, node):
        """
        Open a VNC tunnel to the provided node and return the VNC url.

        :param node: Node to open the VNC tunnel to.
        :type node: :class:`libcloud.compute.base.Node`

        :return: URL of the opened VNC tunnel.
        :rtype: ``str``
        """
        path = "/servers/%s/action/" % (node.id)
        response = self._perform_action(path=path, action="open_vnc", method="POST").object
        vnc_url = response["vnc_url"]

        return vnc_url

    def ex_close_vnc_tunnel(self, node):
        """
        Close a VNC server to the provided node.

        :param node: Node to close the VNC tunnel to.
        :type node: :class:`libcloud.compute.base.Node`

        :return: ``True`` on success, ``False`` otherwise.
        :rtype: ``bool``
        """
        path = "/servers/%s/action/" % (node.id)
        response = self._perform_action(path=path, action="close_vnc", method="POST")

        return response.status == httplib.ACCEPTED

    # Drive extension methods

    def ex_list_library_drives(self):
        """
        Return a list of all the available library drives (pre-installed and
        installation CDs).

        :rtype: ``list`` of :class:`.CloudSigmaDrive` objects
        """
        response = self.connection.request(action="/libdrives/").object
        drives = [self._to_drive(data=item) for item in response["objects"]]

        return drives

    def ex_list_user_drives(self):
        """
        Return a list of all the available user's drives.

        :rtype: ``list`` of :class:`.CloudSigmaDrive` objects
        """
        response = self.connection.request(action="/drives/detail/").object
        drives = [self._to_drive(data=item) for item in response["objects"]]

        return drives

    def list_volumes(self):
        return self.ex_list_user_drives()

    def ex_create_drive(self, name, size, media="disk", ex_avoid=None):
        """
        Create a new drive.

        :param name: Drive name.
        :type name: ``str``

        :param size: Drive size in GBs.
        :type size: ``int``

        :param media: Drive media type (cdrom, disk).
        :type media: ``str``

        :param ex_avoid: A list of other drive uuids to avoid when
                         creating this drive. If provided, drive will
                         attempt to be created on a different
                         physical infrastructure from other drives
                         specified using this argument. (optional)
        :type ex_avoid: ``list``

        :return: Created drive object.
        :rtype: :class:`.CloudSigmaDrive`
        """
        params = {}
        data = {"name": name, "size": size * 1024 * 1024 * 1024, "media": media}

        if ex_avoid:
            params["avoid"] = ",".join(ex_avoid)

        action = "/drives/"
        response = self.connection.request(
            action=action, method="POST", params=params, data=data
        ).object
        drive = self._to_drive(data=response["objects"][0])

        return drive

    def create_volume(self, name, size, media="disk", ex_avoid=None):
        return self.ex_create_drive(name=name, size=size, media=media, ex_avoid=ex_avoid)

    def ex_clone_drive(self, drive, name=None, ex_avoid=None):
        """
        Clone a library or a standard drive.

        :param drive: Drive to clone.
        :type drive: :class:`libcloud.compute.base.NodeImage` or
                     :class:`.CloudSigmaDrive`

        :param name: Optional name for the cloned drive.
        :type name: ``str``

        :param ex_avoid: A list of other drive uuids to avoid when
                         creating this drive. If provided, drive will
                         attempt to be created on a different
                         physical infrastructure from other drives
                         specified using this argument. (optional)
        :type ex_avoid: ``list``

        :return: New cloned drive.
        :rtype: :class:`.CloudSigmaDrive`
        """
        params = {}
        data = {}

        if ex_avoid:
            params["avoid"] = ",".join(ex_avoid)

        if name:
            data["name"] = name

        path = "/drives/%s/action/" % (drive.id)
        response = self._perform_action(
            path=path, action="clone", params=params, data=data, method="POST"
        )
        drive = self._to_drive(data=response.object["objects"][0])

        return drive

    def ex_resize_drive(self, drive, size):
        """
        Resize a drive.

        :param drive: Drive to resize.

        :param size: New drive size in GBs.
        :type size: ``int``

        :return: Drive object which is being resized.
        :rtype: :class:`.CloudSigmaDrive`
        """
        path = "/drives/%s/action/" % (drive.id)
        data = {"name": drive.name, "size": size * 1024 * 1024 * 1024, "media": "disk"}
        response = self._perform_action(path=path, action="resize", method="POST", data=data)

        drive = self._to_drive(data=response.object["objects"][0])

        return drive

    def ex_attach_drive(self, node, drive):
        """
        Attach drive to node

        :param node: Node to attach the drive to.
        :type node: :class:`libcloud.compute.base.Node`

        :param drive: Drive to attach.
        :type drive: :class:`.CloudSigmaDrive`

        :return: ``True`` on success, ``False`` otherwise.
        :rtype: ``bool``
        """
        data = self.ex_get_node(node.id, return_json=True)
        # find the first available device channel to attach the drive
        # device channel format is: {controller}:{unit}
        # total of 920 virtio drives are supported
        # 0:0, …, 0:3, …, 1:0, …, 1:3, …
        # https://cloudsigma-docs.readthedocs.io/en/2.14.3/servers_kvm.html?highlight=dev%20channel#device-channel
        dev_channels = [item["dev_channel"] for item in data["drives"]]
        dev_channel = None

        for controller in range(MAX_VIRTIO_CONTROLLERS):
            for unit in range(MAX_VIRTIO_UNITS):
                if "{}:{}".format(controller, unit) not in dev_channels:
                    dev_channel = "{}:{}".format(controller, unit)

                    break

            if dev_channel:
                break
        else:
            raise CloudSigmaException("Could not attach drive to %s" % (node.id))

        item = {
            "boot_order": None,
            "dev_channel": dev_channel,
            "device": "virtio",
            "drive": str(drive.id),
        }
        data["drives"].append(item)
        action = "/servers/%s/" % (node.id)
        response = self.connection.request(action=action, data=data, method="PUT")

        return response.status == 200

    def attach_volume(self, node, volume):
        return self.ex_attach_drive(node=node, drive=volume)

    def ex_detach_drive(self, node, drive):
        data = self.ex_get_node(node.id, return_json=True)
        data["drives"] = [item for item in data["drives"] if item["drive"]["uuid"] != drive.id]
        action = "/servers/%s/" % (node.id)
        response = self.connection.request(action=action, data=data, method="PUT")

        return response.status == 200

    def detach_volume(self, node, volume):
        return self.ex_detach_drive(node=node, drive=volume)

    def ex_get_drive(self, drive_id):
        """
        Retrieve information about a single drive.

        :param drive_id: ID of the drive to retrieve.
        :type drive_id: ``str``

        :return: Drive object.
        :rtype: :class:`.CloudSigmaDrive`
        """
        action = "/drives/%s/" % (drive_id)
        response = self.connection.request(action=action).object
        drive = self._to_drive(data=response)

        return drive

    def ex_destroy_drive(self, drive):
        action = "/drives/%s/" % (drive.id)
        response = self.connection.request(action=action, method="DELETE")

        return response.status == httplib.NO_CONTENT

    def destroy_volume(self, drive):
        return self.ex_destroy_drive(drive=drive)

    # Firewall policies extension methods

    def ex_list_firewall_policies(self):
        """
        List firewall policies.

        :rtype: ``list`` of :class:`.CloudSigmaFirewallPolicy`
        """
        action = "/fwpolicies/detail/"
        response = self.connection.request(action=action, method="GET").object
        policies = [self._to_firewall_policy(data=item) for item in response["objects"]]

        return policies

    def ex_create_firewall_policy(self, name, rules=None):
        """
        Create a firewall policy.

        :param name: Policy name.
        :type name: ``str``

        :param rules: List of firewall policy rules to associate with this
                      policy. (optional)
        :type rules: ``list`` of ``dict``

        :return: Created firewall policy object.
        :rtype: :class:`.CloudSigmaFirewallPolicy`
        """
        data = {}
        obj = {}
        obj["name"] = name

        if rules:
            obj["rules"] = rules

        data["objects"] = [obj]

        action = "/fwpolicies/"
        response = self.connection.request(action=action, method="POST", data=data).object
        policy = self._to_firewall_policy(data=response["objects"][0])

        return policy

    def ex_attach_firewall_policy(self, policy, node, nic_mac=None):
        """
        Attach firewall policy to a public NIC interface on the server.

        :param policy: Firewall policy to attach.
        :type policy: :class:`.CloudSigmaFirewallPolicy`

        :param node: Node to attach policy to.
        :type node: :class:`libcloud.compute.base.Node`

        :param nic_mac: Optional MAC address of the NIC to add the policy to.
                        If not specified, first public interface is used
                        instead.
        :type nic_mac: ``str``

        :return: Node object to which the policy was attached to.
        :rtype: :class:`libcloud.compute.base.Node`
        """
        nics = copy.deepcopy(node.extra.get("nics", []))

        if nic_mac:
            nic = [n for n in nics if n["mac"] == nic_mac]
        else:
            nic = nics

        if len(nic) == 0:
            raise ValueError("Cannot find the NIC interface to attach " "a policy to")

        nic = nic[0]
        nic["firewall_policy"] = policy.id

        params = {"nics": nics}
        node = self.ex_edit_node(node=node, params=params)

        return node

    def ex_delete_firewall_policy(self, policy):
        """
        Delete a firewall policy.

        :param policy: Policy to delete to.
        :type policy: :class:`.CloudSigmaFirewallPolicy`

        :return: ``True`` on success, ``False`` otherwise.
        :rtype: ``bool``
        """
        action = "/fwpolicies/%s/" % (policy.id)
        response = self.connection.request(action=action, method="DELETE")

        return response.status == httplib.NO_CONTENT

    # Availability groups extension methods

    def ex_list_servers_availability_groups(self):
        """
        Return which running servers share the same physical compute host.

        :return: A list of server UUIDs which share the same physical compute
                 host. Servers which share the same host will be stored under
                 the same list index.
        :rtype: ``list`` of ``list``
        """
        action = "/servers/availability_groups/"
        response = self.connection.request(action=action, method="GET")

        return response.object

    def ex_list_drives_availability_groups(self):
        """
        Return which drives share the same physical storage host.

        :return: A list of drive UUIDs which share the same physical storage
                 host. Drives which share the same host will be stored under
                 the same list index.
        :rtype: ``list`` of ``list``
        """
        action = "/drives/availability_groups/"
        response = self.connection.request(action=action, method="GET")

        return response.object

    # Tag extension methods

    def ex_list_tags(self):
        """
        List all the available tags.

        :rtype: ``list`` of :class:`.CloudSigmaTag` objects
        """
        action = "/tags/detail/"
        response = self.connection.request(action=action, method="GET").object
        tags = [self._to_tag(data=item) for item in response["objects"]]

        return tags

    def ex_get_tag(self, tag_id):
        """
        Retrieve a single tag.

        :param tag_id: ID of the tag to retrieve.
        :type tag_id: ``str``

        :rtype: ``list`` of :class:`.CloudSigmaTag` objects
        """
        action = "/tags/%s/" % (tag_id)
        response = self.connection.request(action=action, method="GET").object
        tag = self._to_tag(data=response)

        return tag

    def ex_create_tag(self, name, resource_uuids=None):
        """
        Create a tag.

        :param name: Tag name.
        :type name: ``str``

        :param resource_uuids: Optional list of resource UUIDs to assign this
                               tag go.
        :type resource_uuids: ``list`` of ``str``

        :return: Created tag object.
        :rtype: :class:`.CloudSigmaTag`
        """
        data = {}
        data["objects"] = [{"name": name}]

        if resource_uuids:
            data["resources"] = resource_uuids

        action = "/tags/"
        response = self.connection.request(action=action, method="POST", data=data).object
        tag = self._to_tag(data=response["objects"][0])

        return tag

    def ex_tag_resource(self, resource, tag):
        """
        Associate tag with the provided resource.

        :param resource: Resource to associate a tag with.
        :type resource: :class:`libcloud.compute.base.Node` or
                        :class:`.CloudSigmaDrive`

        :param tag: Tag to associate with the resources.
        :type tag: :class:`.CloudSigmaTag`

        :return: Updated tag object.
        :rtype: :class:`.CloudSigmaTag`
        """

        if not hasattr(resource, "id"):
            raise ValueError("Resource doesn't have id attribute")

        return self.ex_tag_resources(resources=[resource], tag=tag)

    def ex_tag_resources(self, resources, tag):
        """
        Associate tag with the provided resources.

        :param resources: Resources to associate a tag with.
        :type resources: ``list`` of :class:`libcloud.compute.base.Node` or
                        :class:`.CloudSigmaDrive`

        :param tag: Tag to associate with the resources.
        :type tag: :class:`.CloudSigmaTag`

        :return: Updated tag object.
        :rtype: :class:`.CloudSigmaTag`
        """

        resources = tag.resources[:]

        for resource in resources:
            if not hasattr(resource, "id"):
                raise ValueError("Resource doesn't have id attribute")

            resources.append(resource.id)

        resources = list(set(resources))

        data = {"name": tag.name, "resources": resources}

        action = "/tags/%s/" % (tag.id)
        response = self.connection.request(action=action, method="PUT", data=data).object
        tag = self._to_tag(data=response)

        return tag

    def ex_delete_tag(self, tag):
        """
        Delete a tag.

        :param tag: Tag to delete.
        :type tag: :class:`.CloudSigmaTag`

        :return: ``True`` on success, ``False`` otherwise.
        :rtype: ``bool``
        """
        action = "/tags/%s/" % (tag.id)
        response = self.connection.request(action=action, method="DELETE")

        return response.status == httplib.NO_CONTENT

    # Account extension methods

    def ex_get_balance(self):
        """
        Retrieve account balance information.

        :return: Dictionary with two items ("balance" and "currency").
        :rtype: ``dict``
        """
        action = "/balance/"
        response = self.connection.request(action=action, method="GET")

        return response.object

    def ex_get_pricing(self):
        """
        Retrieve pricing information that are applicable to the cloud.

        :return: Dictionary with pricing information.
        :rtype: ``dict``
        """
        action = "/pricing/"
        response = self.connection.request(action=action, method="GET")

        return response.object

    def ex_get_usage(self):
        """
        Retrieve account current usage information.

        :return: Dictionary with two items ("balance" and "usage").
        :rtype: ``dict``
        """
        action = "/currentusage/"
        response = self.connection.request(action=action, method="GET")

        return response.object

    def ex_list_subscriptions(self, status="all", resources=None):
        """
        List subscriptions for this account.

        :param status: Only return subscriptions with the provided status
                       (optional).
        :type status: ``str``
        :param resources: Only return subscriptions for the provided resources
                          (optional).
        :type resources: ``list``

        :rtype: ``list``
        """
        params = {}

        if status:
            params["status"] = status

        if resources:
            params["resource"] = ",".join(resources)

        response = self.connection.request(action="/subscriptions/", params=params).object
        subscriptions = self._to_subscriptions(data=response)

        return subscriptions

    def ex_toggle_subscription_auto_renew(self, subscription):
        """
        Toggle subscription auto renew status.

        :param subscription: Subscription to toggle the auto renew flag for.
        :type subscription: :class:`.CloudSigmaSubscription`

        :return: ``True`` on success, ``False`` otherwise.
        :rtype: ``bool``
        """
        path = "/subscriptions/%s/action/" % (subscription.id)
        response = self._perform_action(path=path, action="auto_renew", method="POST")

        return response.status == httplib.OK

    def ex_create_subscription(self, amount, period, resource, auto_renew=False):
        """
        Create a new subscription.

        :param amount: Subscription amount. For example, in dssd case this
                       would be disk size in gigabytes.
        :type amount: ``int``

        :param period: Subscription period. For example: 30 days, 1 week, 1
                                            month, ...
        :type period: ``str``

        :param resource: Resource the purchase the subscription for.
        :type resource: ``str``

        :param auto_renew: True to automatically renew the subscription.
        :type auto_renew: ``bool``
        """
        data = [
            {
                "amount": amount,
                "period": period,
                "auto_renew": auto_renew,
                "resource": resource,
            }
        ]

        response = self.connection.request(action="/subscriptions/", data=data, method="POST")
        data = response.object["objects"][0]
        subscription = self._to_subscription(data=data)

        return subscription

    # Misc extension methods

    def ex_list_capabilities(self):
        """
        Retrieve all the basic and sensible limits of the API.

        :rtype: ``dict``
        """
        action = "/capabilities/"
        response = self.connection.request(action=action, method="GET")
        capabilities = response.object

        return capabilities

    def list_key_pairs(self):
        """
        List all the available key pair objects.

        :rtype: ``list`` of :class:`KeyPair` objects
        """
        action = "/keypairs"
        response = self.connection.request(action=action, method="GET").object

        keys = [self._to_key_pair(data=item) for item in response["objects"]]

        return keys

    def get_key_pair(self, key_uuid):
        """
        Retrieve a single key pair.

        :param name: The uuid of the key pair to retrieve.
        :type name: ``str``

        :rtype: :class:`.KeyPair`
        """
        action = "/keypairs/%s/" % (key_uuid)
        response = self.connection.request(action=action, method="GET").object

        return self._to_key_pair(response)

    def create_key_pair(self, name):
        """
        Create a new SSH key.

        :param name: Key pair name.
        :type name: ``str``
        """

        action = "/keypairs/"
        data = {"objects": [{"name": name}]}
        response = self.connection.request(action=action, method="POST", data=data).object

        return self._to_key_pair(response["objects"][0])

    def import_key_pair_from_string(self, name, key_material):
        """
        Import a new public key from string.

        :param name: Key pair name.
        :type name: ``str``

        :param key_material: Public key material.
        :type key_material: ``str``

        :rtype: :class:`.KeyPair` object
        """
        action = "/keypairs/"
        data = {"objects": [{"name": name, "public_key": key_material.replace("\n", "")}]}
        response = self.connection.request(action=action, method="POST", data=data).object

        return self._to_key_pair(response["objects"][0])

    def delete_key_pair(self, key_pair):
        """
        Delete an existing key pair.

        :param key_pair: Key pair object
        :type key_pair: :class:`.KeyPair`

        :rtype: ``bool``
        """
        action = "/keypairs/%s/" % (key_pair.extra["uuid"])
        response = self.connection.request(action=action, method="DELETE")

        return response.status == 204

    def _parse_ips_from_nic(self, nic):
        """
        Parse private and public IP addresses from the provided network
        interface object.

        :param nic: NIC object.
        :type nic: ``dict``

        :return: (public_ips, private_ips) tuple.
        :rtype: ``tuple``
        """
        public_ips, private_ips = [], []

        ipv4_conf = nic["ip_v4_conf"]
        ipv6_conf = nic["ip_v6_conf"]

        ip_v4 = ipv4_conf["ip"] if ipv4_conf else None
        ip_v6 = ipv6_conf["ip"] if ipv6_conf else None

        ipv4 = ip_v4["uuid"] if ip_v4 else None
        ipv6 = ip_v4["uuid"] if ip_v6 else None

        ips = []

        if ipv4:
            ips.append(ipv4)

        if ipv6:
            ips.append(ipv6)

        runtime = nic["runtime"]

        ip_v4 = runtime["ip_v4"] if nic["runtime"] else None
        ip_v6 = runtime["ip_v6"] if nic["runtime"] else None

        ipv4 = ip_v4["uuid"] if ip_v4 else None
        ipv6 = ip_v4["uuid"] if ip_v6 else None

        if ipv4:
            ips.append(ipv4)

        if ipv6:
            ips.append(ipv6)

        ips = set(ips)

        for ip in ips:
            if is_private_subnet(ip):
                private_ips.append(ip)
            else:
                public_ips.append(ip)

        return public_ips, private_ips

    def _to_node(self, data):
        id = data["uuid"]
        name = data["name"]
        state = self.NODE_STATE_MAP.get(data["status"], NodeState.UNKNOWN)

        public_ips = []
        private_ips = []
        extra = {
            "cpus": data["cpu"] / 2000,
            "memory": data["mem"] / 1024 / 1024,
            "nics": data["nics"],
            "vnc_password": data["vnc_password"],
            "meta": data["meta"],
            "runtime": data["runtime"],
            "drives": data["drives"],
        }
        # find image name and boot drive size
        image = None
        drive_size = 0

        for item in extra["drives"]:
            if item["boot_order"] == 1:
                drive = self.ex_get_drive(item["drive"]["uuid"])
                drive_size = drive.size
                image = "{} {}".format(
                    drive.extra.get("distribution", ""), drive.extra.get("version", "")
                )

                break
        # try to find if node size is from example sizes given by CloudSigma
        try:
            kwargs = SPECS_TO_SIZE[(extra["cpus"], extra["memory"], drive_size)]
            size = CloudSigmaNodeSize(**kwargs, driver=self)
        except KeyError:
            id_to_hash = str(extra["cpus"]) + str(extra["memory"]) + str(drive_size)
            size_id = hashlib.md5(id_to_hash.encode("utf-8")).hexdigest()  # nosec
            size_name = "custom, {} CPUs, {}MB RAM, {}GB disk".format(
                extra["cpus"], extra["memory"], drive_size
            )  # noqa
            size = CloudSigmaNodeSize(
                id=size_id,
                name=size_name,
                cpu=extra["cpus"],
                ram=extra["memory"],
                disk=drive_size,
                bandwidth=None,
                price=0,
                driver=self,
            )

        for nic in data["nics"]:
            _public_ips, _private_ips = self._parse_ips_from_nic(nic=nic)

            public_ips.extend(_public_ips)
            private_ips.extend(_private_ips)

        node = Node(
            id=id,
            name=name,
            state=state,
            public_ips=public_ips,
            image=image,
            private_ips=private_ips,
            driver=self,
            size=size,
            extra=extra,
        )

        return node

    def _to_image(self, data):
        extra_keys = [
            "description",
            "arch",
            "image_type",
            "os",
            "licenses",
            "media",
            "meta",
        ]

        id = data["uuid"]
        name = data["name"]
        extra = self._extract_values(obj=data, keys=extra_keys)

        image = NodeImage(id=id, name=name, driver=self, extra=extra)

        return image

    def _to_drive(self, data):
        id = data["uuid"]
        name = data["name"]
        size = data["size"] / 1024 / 1024 / 1024
        media = data["media"]
        status = data["status"]
        extra = {
            "mounted_on": data.get("mounted_on", []),
            "storage_type": data.get("storage_type", ""),
            "distribution": data["meta"].get("distribution", ""),
            "version": data["meta"].get("version", ""),
            "os": data["meta"].get("os", ""),
            "paid": data["meta"].get("paid", ""),
            "architecture": data["meta"].get("arch", ""),
            "created_at": data["meta"].get("created_at", ""),
        }

        drive = CloudSigmaDrive(
            id=id,
            name=name,
            size=size,
            media=media,
            status=status,
            driver=self,
            extra=extra,
        )

        return drive

    def _to_tag(self, data):
        resources = data["resources"]
        resources = [resource["uuid"] for resource in resources]

        tag = CloudSigmaTag(id=data["uuid"], name=data["name"], resources=resources)

        return tag

    def _to_subscriptions(self, data):
        subscriptions = []

        for item in data["objects"]:
            subscription = self._to_subscription(data=item)
            subscriptions.append(subscription)

        return subscriptions

    def _to_subscription(self, data):
        if data.get("start_time", None):
            start_time = parse_date(data["start_time"])
        else:
            start_time = None

        if data.get("end_time", None):
            end_time = parse_date(data["end_time"])
        else:
            end_time = None

        obj_uuid = data["subscribed_object"]

        subscription = CloudSigmaSubscription(
            id=data["id"],
            resource=data["resource"],
            amount=int(data["amount"]),
            period=data["period"],
            status=data["status"],
            price=data["price"],
            start_time=start_time,
            end_time=end_time,
            auto_renew=data["auto_renew"],
            subscribed_object=obj_uuid,
        )

        return subscription

    def _to_firewall_policy(self, data):
        rules = []

        for item in data.get("rules", []):
            rule = CloudSigmaFirewallPolicyRule(
                action=item["action"],
                direction=item["direction"],
                ip_proto=item["ip_proto"],
                src_ip=item["src_ip"],
                src_port=item["src_port"],
                dst_ip=item["dst_ip"],
                dst_port=item["dst_port"],
                comment=item["comment"],
            )
            rules.append(rule)

        policy = CloudSigmaFirewallPolicy(id=data["uuid"], name=data["name"], rules=rules)

        return policy

    def _to_key_pair(self, data):
        extra = {
            "uuid": data["uuid"],
            "tags": data["tags"],
            "resource_uri": data["resource_uri"],
            "permissions": data["permissions"],
            "meta": data["meta"],
        }

        return KeyPair(
            name=data["name"],
            public_key=data["public_key"],
            fingerprint=data["fingerprint"],
            driver=self,
            private_key=data["private_key"],
            extra=extra,
        )

    def _perform_action(self, path, action, method="POST", params=None, data=None):
        """
        Perform API action and return response object.
        """

        if params:
            params = params.copy()
        else:
            params = {}

        params["do"] = action
        response = self.connection.request(action=path, method=method, params=params, data=data)

        return response

    def _is_installation_cd(self, image):
        """
        Detect if the provided image is an installation CD.

        :rtype: ``bool``
        """

        if isinstance(image, CloudSigmaDrive) and image.media == "cdrom":
            return True

        return False

    def _extract_values(self, obj, keys):
        """
        Extract values from a dictionary and return a new dictionary with
        extracted values.

        :param obj: Dictionary to extract values from.
        :type obj: ``dict``

        :param keys: Keys to extract.
        :type keys: ``list``

        :return: Dictionary with extracted values.
        :rtype: ``dict``
        """
        result = {}

        for key in keys:
            result[key] = obj[key]

        return result

    def _wait_for_drive_state_transition(self, drive, state, timeout=DRIVE_TRANSITION_TIMEOUT):
        """
        Wait for a drive to transition to the provided state.

        Note: This function blocks and periodically calls "GET drive" endpoint
        to check if the drive has already transitioned to the desired state.

        :param drive: Drive to wait for.
        :type drive: :class:`.CloudSigmaDrive`

        :param state: Desired drive state.
        :type state: ``str``

        :param timeout: How long to wait for the transition (in seconds) before
                        timing out.
        :type timeout: ``int``

        :return: Drive object.
        :rtype: :class:`.CloudSigmaDrive`
        """

        start_time = time.time()

        while drive.status != state:
            drive = self.ex_get_drive(drive_id=drive.id)

            if drive.status == state:
                break

            current_time = time.time()
            delta = current_time - start_time

            if delta >= timeout:
                msg = "Timed out while waiting for drive transition " "(timeout=%s seconds)" % (
                    timeout
                )
                raise Exception(msg)

            time.sleep(self.DRIVE_TRANSITION_SLEEP_INTERVAL)

        return drive

    def _ex_connection_class_kwargs(self):
        """
        Return the host value based on the user supplied region.
        """
        kwargs = {}

        if not self._host_argument_set:
            kwargs["host"] = API_ENDPOINTS_2_0[self.region]["host"]

        return kwargs



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Vultr Driver
"""

import json
import time
import base64
from typing import Any, Dict, List, Union, Optional
from functools import update_wrapper

from libcloud.utils.py3 import httplib, urlencode
from libcloud.common.base import JsonResponse, ConnectionKey
from libcloud.common.types import LibcloudError, InvalidCredsError, ServiceUnavailableError
from libcloud.common.vultr import (
    DEFAULT_API_VERSION,
    VultrNetwork,
    VultrConnectionV2,
    VultrNodeSnapshot,
)
from libcloud.compute.base import (
    Node,
    KeyPair,
    NodeSize,
    NodeImage,
    NodeDriver,
    NodeLocation,
    StorageVolume,
)
from libcloud.compute.types import Provider, NodeState, StorageVolumeState, VolumeSnapshotState
from libcloud.utils.iso8601 import parse_date
from libcloud.utils.publickey import get_pubkey_openssh_fingerprint

# For matching region by id
VULTR_COMPUTE_INSTANCE_LOCATIONS = {
    "1": {
        "DCID": "1",
        "name": "New Jersey",
        "country": "US",
        "continent": "North America",
        "state": "NJ",
        "regioncode": "EWR",
    },
    "2": {
        "DCID": "2",
        "name": "Chicago",
        "country": "US",
        "continent": "North America",
        "state": "IL",
        "regioncode": "ORD",
    },
    "3": {
        "DCID": "3",
        "name": "Dallas",
        "country": "US",
        "continent": "North America",
        "state": "TX",
        "regioncode": "DFW",
    },
    "4": {
        "DCID": "4",
        "name": "Seattle",
        "country": "US",
        "continent": "North America",
        "state": "WA",
        "regioncode": "SEA",
    },
    "5": {
        "DCID": "5",
        "name": "Los Angeles",
        "country": "US",
        "continent": "North America",
        "state": "CA",
        "regioncode": "LAX",
    },
    "6": {
        "DCID": "6",
        "name": "Atlanta",
        "country": "US",
        "continent": "North America",
        "state": "GA",
        "regioncode": "ATL",
    },
    "7": {
        "DCID": "7",
        "name": "Amsterdam",
        "country": "NL",
        "continent": "Europe",
        "state": "",
        "regioncode": "AMS",
    },
    "8": {
        "DCID": "8",
        "name": "London",
        "country": "GB",
        "continent": "Europe",
        "state": "",
        "regioncode": "LHR",
    },
    "9": {
        "DCID": "9",
        "name": "Frankfurt",
        "country": "DE",
        "continent": "Europe",
        "state": "",
        "regioncode": "FRA",
    },
    "12": {
        "DCID": "12",
        "name": "Silicon Valley",
        "country": "US",
        "continent": "North America",
        "state": "CA",
        "regioncode": "SJC",
    },
    "19": {
        "DCID": "19",
        "name": "Sydney",
        "country": "AU",
        "continent": "Australia",
        "state": "",
        "regioncode": "SYD",
    },
    "22": {
        "DCID": "22",
        "name": "Toronto",
        "country": "CA",
        "continent": "North America",
        "state": "",
        "regioncode": "YTO",
    },
    "24": {
        "DCID": "24",
        "name": "Paris",
        "country": "FR",
        "continent": "Europe",
        "state": "",
        "regioncode": "CDG",
    },
    "25": {
        "DCID": "25",
        "name": "Tokyo",
        "country": "JP",
        "continent": "Asia",
        "state": "",
        "regioncode": "NRT",
    },
    "34": {
        "DCID": "34",
        "name": "Seoul",
        "country": "KR",
        "continent": "Asia",
        "state": "",
        "regioncode": "ICN",
    },
    "39": {
        "DCID": "39",
        "name": "Miami",
        "country": "US",
        "continent": "North America",
        "state": "FL",
        "regioncode": "MIA",
    },
    "40": {
        "DCID": "40",
        "name": "Singapore",
        "country": "SG",
        "continent": "Asia",
        "state": "",
        "regioncode": "SGP",
    },
}
# For matching image by id
VULTR_COMPUTE_INSTANCE_IMAGES = {
    "127": {
        "OSID": 127,
        "name": "CentOS 6 x64",
        "arch": "x64",
        "family": "centos",
        "windows": False,
    },
    "147": {
        "OSID": 147,
        "name": "CentOS 6 i386",
        "arch": "i386",
        "family": "centos",
        "windows": False,
    },
    "167": {
        "OSID": 167,
        "name": "CentOS 7 x64",
        "arch": "x64",
        "family": "centos",
        "windows": False,
    },
    "381": {
        "OSID": 381,
        "name": "CentOS 7 SELinux x64",
        "arch": "x64",
        "family": "centos",
        "windows": False,
    },
    "362": {
        "OSID": 362,
        "name": "CentOS 8 x64",
        "arch": "x64",
        "family": "centos",
        "windows": False,
    },
    "401": {
        "OSID": 401,
        "name": "CentOS 8 Stream x64",
        "arch": "x64",
        "family": "centos",
        "windows": False,
    },
    "215": {
        "OSID": 215,
        "name": "Ubuntu 16.04 x64",
        "arch": "x64",
        "family": "ubuntu",
        "windows": False,
    },
    "216": {
        "OSID": 216,
        "name": "Ubuntu 16.04 i386",
        "arch": "i386",
        "family": "ubuntu",
        "windows": False,
    },
    "270": {
        "OSID": 270,
        "name": "Ubuntu 18.04 x64",
        "arch": "x64",
        "family": "ubuntu",
        "windows": False,
    },
    "387": {
        "OSID": 387,
        "name": "Ubuntu 20.04 x64",
        "arch": "x64",
        "family": "ubuntu",
        "windows": False,
    },
    "194": {
        "OSID": 194,
        "name": "Debian 8 i386 (jessie)",
        "arch": "i386",
        "family": "debian",
        "windows": False,
    },
    "244": {
        "OSID": 244,
        "name": "Debian 9 x64 (stretch)",
        "arch": "x64",
        "family": "debian",
        "windows": False,
    },
    "352": {
        "OSID": 352,
        "name": "Debian 10 x64 (buster)",
        "arch": "x64",
        "family": "debian",
        "windows": False,
    },
    "230": {
        "OSID": 230,
        "name": "FreeBSD 11 x64",
        "arch": "x64",
        "family": "freebsd",
        "windows": False,
    },
    "327": {
        "OSID": 327,
        "name": "FreeBSD 12 x64",
        "arch": "x64",
        "family": "freebsd",
        "windows": False,
    },
    "366": {
        "OSID": 366,
        "name": "OpenBSD 6.6 x64",
        "arch": "x64",
        "family": "openbsd",
        "windows": False,
    },
    "394": {
        "OSID": 394,
        "name": "OpenBSD 6.7 x64",
        "arch": "x64",
        "family": "openbsd",
        "windows": False,
    },
    "391": {
        "OSID": 391,
        "name": "Fedora CoreOS",
        "arch": "x64",
        "family": "fedora-coreos",
        "windows": False,
    },
    "367": {
        "OSID": 367,
        "name": "Fedora 31 x64",
        "arch": "x64",
        "family": "fedora",
        "windows": False,
    },
    "389": {
        "OSID": 389,
        "name": "Fedora 32 x64",
        "arch": "x64",
        "family": "fedora",
        "windows": False,
    },
    "124": {
        "OSID": 124,
        "name": "Windows 2012 R2 x64",
        "arch": "x64",
        "family": "windows",
        "windows": False,
    },
    "240": {
        "OSID": 240,
        "name": "Windows 2016 x64",
        "arch": "x64",
        "family": "windows",
        "windows": False,
    },
    "159": {
        "OSID": 159,
        "name": "Custom",
        "arch": "x64",
        "family": "iso",
        "windows": False,
    },
    "164": {
        "OSID": 164,
        "name": "Snapshot",
        "arch": "x64",
        "family": "snapshot",
        "windows": False,
    },
    "180": {
        "OSID": 180,
        "name": "Backup",
        "arch": "x64",
        "family": "backup",
        "windows": False,
    },
    "186": {
        "OSID": 186,
        "name": "Application",
        "arch": "x64",
        "family": "application",
        "windows": False,
    },
}
VULTR_COMPUTE_INSTANCE_SIZES = {
    "201": {
        "VPSPLANID": "201",
        "name": "1024 MB RAM,25 GB SSD,1.00 TB BW",
        "vcpu_count": "1",
        "ram": "1024",
        "disk": "25",
        "bandwidth": "1.00",
        "bandwidth_gb": "1024",
        "price_per_month": "5.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "202": {
        "VPSPLANID": "202",
        "name": "2048 MB RAM,55 GB SSD,2.00 TB BW",
        "vcpu_count": "1",
        "ram": "2048",
        "disk": "55",
        "bandwidth": "2.00",
        "bandwidth_gb": "2048",
        "price_per_month": "10.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "203": {
        "VPSPLANID": "203",
        "name": "4096 MB RAM,80 GB SSD,3.00 TB BW",
        "vcpu_count": "2",
        "ram": "4096",
        "disk": "80",
        "bandwidth": "3.00",
        "bandwidth_gb": "3072",
        "price_per_month": "20.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "204": {
        "VPSPLANID": "204",
        "name": "8192 MB RAM,160 GB SSD,4.00 TB BW",
        "vcpu_count": "4",
        "ram": "8192",
        "disk": "160",
        "bandwidth": "4.00",
        "bandwidth_gb": "4096",
        "price_per_month": "40.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "205": {
        "VPSPLANID": "205",
        "name": "16384 MB RAM,320 GB SSD,5.00 TB BW",
        "vcpu_count": "6",
        "ram": "16384",
        "disk": "320",
        "bandwidth": "5.00",
        "bandwidth_gb": "5120",
        "price_per_month": "80.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "206": {
        "VPSPLANID": "206",
        "name": "32768 MB RAM,640 GB SSD,6.00 TB BW",
        "vcpu_count": "8",
        "ram": "32768",
        "disk": "640",
        "bandwidth": "6.00",
        "bandwidth_gb": "6144",
        "price_per_month": "160.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "207": {
        "VPSPLANID": "207",
        "name": "65536 MB RAM,1280 GB SSD,10.00 TB BW",
        "vcpu_count": "16",
        "ram": "65536",
        "disk": "1280",
        "bandwidth": "10.00",
        "bandwidth_gb": "10240",
        "price_per_month": "320.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "208": {
        "VPSPLANID": "208",
        "name": "98304 MB RAM,1600 GB SSD,15.00 TB BW",
        "vcpu_count": "24",
        "ram": "98304",
        "disk": "1600",
        "bandwidth": "15.00",
        "bandwidth_gb": "15360",
        "price_per_month": "640.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "115": {
        "VPSPLANID": "115",
        "name": "8192 MB RAM,110 GB SSD,10.00 TB BW",
        "vcpu_count": "2",
        "ram": "8192",
        "disk": "110",
        "bandwidth": "10.00",
        "bandwidth_gb": "10240",
        "price_per_month": "60.00",
        "plan_type": "DEDICATED",
        "windows": False,
    },
    "116": {
        "VPSPLANID": "116",
        "name": "16384 MB RAM,2x110 GB SSD,20.00 TB BW",
        "vcpu_count": "4",
        "ram": "16384",
        "disk": "110",
        "bandwidth": "20.00",
        "bandwidth_gb": "20480",
        "price_per_month": "120.00",
        "plan_type": "DEDICATED",
        "windows": False,
    },
    "117": {
        "VPSPLANID": "117",
        "name": "24576 MB RAM,3x110 GB SSD,30.00 TB BW",
        "vcpu_count": "6",
        "ram": "24576",
        "disk": "110",
        "bandwidth": "30.00",
        "bandwidth_gb": "30720",
        "price_per_month": "180.00",
        "plan_type": "DEDICATED",
        "windows": False,
    },
    "118": {
        "VPSPLANID": "118",
        "name": "32768 MB RAM,4x110 GB SSD,40.00 TB BW",
        "vcpu_count": "8",
        "ram": "32768",
        "disk": "110",
        "bandwidth": "40.00",
        "bandwidth_gb": "40960",
        "price_per_month": "240.00",
        "plan_type": "DEDICATED",
        "windows": False,
    },
    "400": {
        "VPSPLANID": "400",
        "name": "1024 MB RAM,32 GB SSD,1.00 TB BW",
        "vcpu_count": "1",
        "ram": "1024",
        "disk": "32",
        "bandwidth": "1.00",
        "bandwidth_gb": "1024",
        "price_per_month": "6.00",
        "plan_type": "HIGHFREQUENCY",
        "windows": False,
    },
    "401": {
        "VPSPLANID": "401",
        "name": "2048 MB RAM,64 GB SSD,2.00 TB BW",
        "vcpu_count": "1",
        "ram": "2048",
        "disk": "64",
        "bandwidth": "2.00",
        "bandwidth_gb": "2048",
        "price_per_month": "12.00",
        "plan_type": "HIGHFREQUENCY",
        "windows": False,
    },
    "402": {
        "VPSPLANID": "402",
        "name": "4096 MB RAM,128 GB SSD,3.00 TB BW",
        "vcpu_count": "2",
        "ram": "4096",
        "disk": "128",
        "bandwidth": "3.00",
        "bandwidth_gb": "3072",
        "price_per_month": "24.00",
        "plan_type": "HIGHFREQUENCY",
        "windows": False,
    },
    "403": {
        "VPSPLANID": "403",
        "name": "8192 MB RAM,256 GB SSD,4.00 TB BW",
        "vcpu_count": "3",
        "ram": "8192",
        "disk": "256",
        "bandwidth": "4.00",
        "bandwidth_gb": "4096",
        "price_per_month": "48.00",
        "plan_type": "HIGHFREQUENCY",
        "windows": False,
    },
    "404": {
        "VPSPLANID": "404",
        "name": "16384 MB RAM,384 GB SSD,5.00 TB BW",
        "vcpu_count": "4",
        "ram": "16384",
        "disk": "384",
        "bandwidth": "5.00",
        "bandwidth_gb": "5120",
        "price_per_month": "96.00",
        "plan_type": "HIGHFREQUENCY",
        "windows": False,
    },
    "405": {
        "VPSPLANID": "405",
        "name": "32768 MB RAM,512 GB SSD,6.00 TB BW",
        "vcpu_count": "8",
        "ram": "32768",
        "disk": "512",
        "bandwidth": "6.00",
        "bandwidth_gb": "6144",
        "price_per_month": "192.00",
        "plan_type": "HIGHFREQUENCY",
        "windows": False,
    },
    "406": {
        "VPSPLANID": "406",
        "name": "49152 MB RAM,768 GB SSD,8.00 TB BW",
        "vcpu_count": "12",
        "ram": "49152",
        "disk": "768",
        "bandwidth": "8.00",
        "bandwidth_gb": "8192",
        "price_per_month": "256.00",
        "plan_type": "HIGHFREQUENCY",
        "windows": False,
    },
}


class rate_limited:
    """
    Decorator for retrying Vultr calls that are rate-limited.

    :param int sleep: Seconds to sleep after being rate-limited.
    :param int retries: Number of retries.
    """

    def __init__(self, sleep=0.5, retries=1):
        self.sleep = sleep
        self.retries = retries

    def __call__(self, call):
        """
        Run ``call`` method until it's not rate-limited.

        The method is invoked while it returns 503 Service Unavailable or the
        allowed number of retries is reached.

        :param callable call: Method to be decorated.
        """

        def wrapper(*args, **kwargs):
            last_exception = None

            for _ in range(self.retries + 1):
                try:
                    return call(*args, **kwargs)
                except ServiceUnavailableError as e:
                    last_exception = e
                    time.sleep(self.sleep)  # hit by rate limit, let's sleep

            if last_exception:
                raise last_exception  # pylint: disable=raising-bad-type

        update_wrapper(wrapper, call)
        return wrapper


class VultrResponse(JsonResponse):
    def parse_error(self):
        if self.status == httplib.OK:
            body = self.parse_body()
            return body
        elif self.status == httplib.FORBIDDEN:
            raise InvalidCredsError(self.body)
        elif self.status == httplib.SERVICE_UNAVAILABLE:
            raise ServiceUnavailableError(self.body)
        else:
            raise LibcloudError(self.body)


class SSHKey:
    def __init__(self, id, name, pub_key):
        self.id = id
        self.name = name
        self.pub_key = pub_key

    def __repr__(self):
        return ("<SSHKey: id=%s, name=%s, pub_key=%s>") % (
            self.id,
            self.name,
            self.pub_key,
        )


class VultrConnection(ConnectionKey):
    """
    Connection class for the Vultr driver.
    """

    host = "api.vultr.com"
    responseCls = VultrResponse
    unauthenticated_endpoints = {  # {action: methods}
        "/v1/app/list": ["GET"],
        "/v1/os/list": ["GET"],
        "/v1/plans/list": ["GET"],
        "/v1/plans/list_vc2": ["GET"],
        "/v1/plans/list_vdc2": ["GET"],
        "/v1/regions/availability": ["GET"],
        "/v1/regions/list": ["GET"],
    }

    def add_default_headers(self, headers):
        """
        Adds ``API-Key`` default header.

        :return: Updated headers.
        :rtype: dict
        """

        if self.require_api_key():
            headers.update({"API-Key": self.key})
        return headers

    def encode_data(self, data):
        return urlencode(data)

    @rate_limited()
    def get(self, url):
        return self.request(url)

    @rate_limited()
    def post(self, url, data):
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        return self.request(url, data=data, headers=headers, method="POST")

    def require_api_key(self):
        """
        Check whether this call (method + action) must be authenticated.

        :return: True if ``API-Key`` header required, False otherwise.
        :rtype: bool
        """

        try:
            return self.method not in self.unauthenticated_endpoints[self.action]
        except KeyError:
            return True


class VultrNodeDriverHelper:
    """
    VultrNode helper class.
    """

    def handle_extra(self, extra_keys, data):
        extra = {}
        for key in extra_keys:
            if key in data:
                extra[key] = data[key]
        return extra


class VultrNodeDriver(NodeDriver):
    type = Provider.VULTR
    name = "Vultr"
    website = "https://www.vultr.com"

    def __new__(
        cls,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        api_version=DEFAULT_API_VERSION,
        region=None,
        **kwargs,
    ):
        if cls is VultrNodeDriver:
            if api_version == "1":
                cls = VultrNodeDriverV1
            elif api_version == "2":
                cls = VultrNodeDriverV2
            else:
                raise NotImplementedError(
                    "No Vultr driver found for API version: %s" % (api_version)
                )
        return super().__new__(cls)


class VultrNodeDriverV1(VultrNodeDriver):
    """
    VultrNode node driver.
    """

    connectionCls = VultrConnection

    NODE_STATE_MAP = {"pending": NodeState.PENDING, "active": NodeState.RUNNING}

    EX_CREATE_YES_NO_ATTRIBUTES = [
        "enable_ipv6",
        "enable_private_network",
        "auto_backups",
        "notify_activate",
        "ddos_protection",
    ]

    EX_CREATE_ID_ATTRIBUTES = {
        "iso_id": "ISOID",
        "script_id": "SCRIPTID",
        "snapshot_id": "SNAPSHOTID",
        "app_id": "APPID",
    }

    EX_CREATE_ATTRIBUTES = [
        "ipxe_chain_url",
        "label",
        "userdata",
        "reserved_ip_v4",
        "hostname",
        "tag",
    ]
    EX_CREATE_ATTRIBUTES.extend(EX_CREATE_YES_NO_ATTRIBUTES)
    EX_CREATE_ATTRIBUTES.extend(EX_CREATE_ID_ATTRIBUTES.keys())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._helper = VultrNodeDriverHelper()

    def list_nodes(self):
        return self._list_resources("/v1/server/list", self._to_node)

    def list_key_pairs(self):
        """
        List all the available SSH keys.
        :return: Available SSH keys.
        :rtype: ``list`` of :class:`SSHKey`
        """
        return self._list_resources("/v1/sshkey/list", self._to_ssh_key)

    def create_key_pair(self, name, public_key=""):
        """
        Create a new SSH key.
        :param name: Name of the new SSH key
        :type name: ``str``

        :key public_key: Public part of the new SSH key
        :type name: ``str``

        :return: True on success
        :rtype: ``bool``
        """
        params = {"name": name, "ssh_key": public_key}
        res = self.connection.post("/v1/sshkey/create", params)
        return res.status == httplib.OK

    def delete_key_pair(self, key_pair):
        """
        Delete an SSH key.
        :param key_pair: The SSH key to delete
        :type key_pair: :class:`SSHKey`

        :return: True on success
        :rtype: ``bool``
        """
        params = {"SSHKEYID": key_pair.id}
        res = self.connection.post("/v1/sshkey/destroy", params)
        return res.status == httplib.OK

    def list_locations(self):
        return self._list_resources("/v1/regions/list", self._to_location)

    def list_sizes(self):
        return self._list_resources("/v1/plans/list", self._to_size)

    def list_images(self):
        return self._list_resources("/v1/os/list", self._to_image)

    # pylint: disable=too-many-locals
    def create_node(self, name, size, image, location, ex_ssh_key_ids=None, ex_create_attr=None):
        """
        Create a node

        :param name: Name for the new node
        :type name: ``str``

        :param size: Size of the new node
        :type size: :class:`NodeSize`

        :param image: Image for the new node
        :type image: :class:`NodeImage`

        :param location: Location of the new node
        :type location: :class:`NodeLocation`

        :param ex_ssh_key_ids: IDs of the SSH keys to initialize
        :type ex_sshkeyid: ``list`` of ``str``

        :param ex_create_attr: Extra attributes for node creation
        :type ex_create_attr: ``dict``

        The `ex_create_attr` parameter can include the following dictionary
        key and value pairs:

        * `ipxe_chain_url`: ``str`` for specifying URL to boot via IPXE
        * `iso_id`: ``str`` the ID of a specific ISO to mount,
          only meaningful with the `Custom` `NodeImage`
        * `script_id`: ``int`` ID of a startup script to execute on boot,
          only meaningful when the `NodeImage` is not `Custom`
        * 'snapshot_id`: ``str`` Snapshot ID to restore for the initial
          installation, only meaningful with the `Snapshot` `NodeImage`
        * `enable_ipv6`: ``bool`` Whether an IPv6 subnet should be assigned
        * `enable_private_network`: ``bool`` Whether private networking
          support should be added
        * `label`: ``str`` Text label to be shown in the control panel
        * `auto_backups`: ``bool`` Whether automatic backups should be enabled
        * `app_id`: ``int`` App ID to launch if launching an application,
          only meaningful when the `NodeImage` is `Application`
        * `userdata`: ``str`` Base64 encoded cloud-init user-data
        * `notify_activate`: ``bool`` Whether an activation email should be
          sent when the server is ready
        * `ddos_protection`: ``bool`` Whether DDOS protection should be enabled
        * `reserved_ip_v4`: ``str`` IP address of the floating IP to use as
          the main IP of this server
        * `hostname`: ``str`` The hostname to assign to this server
        * `tag`: ``str`` The tag to assign to this server

        :return: The newly created node.
        :rtype: :class:`Node`

        """
        params = {
            "DCID": location.id,
            "VPSPLANID": size.id,
            "OSID": image.id,
            "label": name,
        }

        if ex_ssh_key_ids is not None:
            params["SSHKEYID"] = ",".join(ex_ssh_key_ids)

        ex_create_attr = ex_create_attr or {}
        for key, value in ex_create_attr.items():
            if key in self.EX_CREATE_ATTRIBUTES:
                if key in self.EX_CREATE_YES_NO_ATTRIBUTES:
                    params[key] = "yes" if value else "no"
                else:
                    if key in self.EX_CREATE_ID_ATTRIBUTES:
                        key = self.EX_CREATE_ID_ATTRIBUTES[key]
                    params[key] = value

        result = self.connection.post("/v1/server/create", params)
        if result.status != httplib.OK:
            return False

        subid = result.object["SUBID"]

        retry_count = 3
        created_node = None

        for _ in range(retry_count):
            try:
                nodes = self.list_nodes()
                created_node = [n for n in nodes if n.id == subid][0]
            except IndexError:
                time.sleep(1)
            else:
                break

        return created_node

    def reboot_node(self, node):
        params = {"SUBID": node.id}
        res = self.connection.post("/v1/server/reboot", params)

        return res.status == httplib.OK

    def destroy_node(self, node):
        params = {"SUBID": node.id}
        res = self.connection.post("/v1/server/destroy", params)

        return res.status == httplib.OK

    def _list_resources(self, url, tranform_func):
        data = self.connection.get(url).object
        sorted_key = sorted(data)
        return [tranform_func(data[key]) for key in sorted_key]

    def _to_node(self, data):
        if "status" in data:
            state = self.NODE_STATE_MAP.get(data["status"], NodeState.UNKNOWN)
            if state == NodeState.RUNNING and data["power_status"] != "running":
                state = NodeState.STOPPED
        else:
            state = NodeState.UNKNOWN

        if "main_ip" in data and data["main_ip"] is not None:
            public_ips = [data["main_ip"]]
        else:
            public_ips = []
        # simple check that we have ip address in value
        if len(data["internal_ip"]) > 0:
            private_ips = [data["internal_ip"]]
        else:
            private_ips = []
        created_at = parse_date(data["date_created"])

        # response ordering
        extra_keys = [
            "location",  # Location name
            "default_password",
            "pending_charges",
            "cost_per_month",
            "current_bandwidth_gb",
            "allowed_bandwidth_gb",
            "netmask_v4",
            "gateway_v4",
            "power_status",
            "server_state",
            "v6_networks",
            # TODO: Does we really need kvm_url?
            "kvm_url",
            "auto_backups",
            "tag",
            # "OSID",  # Operating system to use. See v1/os/list.
            "APPID",
            "FIREWALLGROUPID",
        ]
        extra = self._helper.handle_extra(extra_keys, data)

        resolve_data = VULTR_COMPUTE_INSTANCE_IMAGES.get(data["OSID"])
        if resolve_data:
            image = self._to_image(resolve_data)
        else:
            image = None

        resolve_data = VULTR_COMPUTE_INSTANCE_SIZES.get(data["VPSPLANID"])
        if resolve_data:
            size = self._to_size(resolve_data)
        else:
            size = None

        # resolve_data = VULTR_COMPUTE_INSTANCE_LOCATIONS.get(data['DCID'])
        # if resolve_data:
        #     location = self._to_location(resolve_data)
        # extra['location'] = location

        node = Node(
            id=data["SUBID"],
            name=data["label"],
            state=state,
            public_ips=public_ips,
            private_ips=private_ips,
            image=image,
            size=size,
            extra=extra,
            created_at=created_at,
            driver=self,
        )

        return node

    def _to_location(self, data):
        extra_keys = [
            "continent",
            "state",
            "ddos_protection",
            "block_storage",
            "regioncode",
        ]
        extra = self._helper.handle_extra(extra_keys, data)

        return NodeLocation(
            id=data["DCID"],
            name=data["name"],
            country=data["country"],
            extra=extra,
            driver=self,
        )

    def _to_size(self, data):
        extra_keys = [
            "vcpu_count",
            "plan_type",
            "available_locations",
        ]
        extra = self._helper.handle_extra(extra_keys, data)

        # backward compatibility
        if extra.get("vcpu_count").isdigit():
            extra["vcpu_count"] = int(extra["vcpu_count"])

        ram = int(data["ram"])
        disk = int(data["disk"])
        # NodeSize accepted int instead float
        bandwidth = int(float(data["bandwidth"]))
        price = float(data["price_per_month"])
        return NodeSize(
            id=data["VPSPLANID"],
            name=data["name"],
            ram=ram,
            disk=disk,
            bandwidth=bandwidth,
            price=price,
            extra=extra,
            driver=self,
        )

    def _to_image(self, data):
        extra_keys = ["arch", "family"]
        extra = self._helper.handle_extra(extra_keys, data)
        return NodeImage(id=data["OSID"], name=data["name"], extra=extra, driver=self)

    def _to_ssh_key(self, data):
        return SSHKey(id=data["SSHKEYID"], name=data["name"], pub_key=data["ssh_key"])


class VultrNodeDriverV2(VultrNodeDriver):
    """
    Vultr API v2 NodeDriver.
    """

    connectionCls = VultrConnectionV2
    NODE_STATE_MAP = {
        "active": NodeState.RUNNING,
        "halted": NodeState.STOPPED,
        "rebooting": NodeState.REBOOTING,
        "resizing": NodeState.RECONFIGURING,
        "pending": NodeState.PENDING,
    }

    VOLUME_STATE_MAP = {
        "active": StorageVolumeState.AVAILABLE,
        "pending": StorageVolumeState.CREATING,
    }

    SNAPSHOT_STATE_MAP = {
        "complete": VolumeSnapshotState.AVAILABLE,
        "pending": VolumeSnapshotState.CREATING,
    }

    def list_nodes(self, ex_list_bare_metals: bool = True) -> List[Node]:
        """List all nodes.

        :keyword ex_list_bare_metals: Whether to fetch bare metal nodes.
        :type    ex_list_bare_metals: ``bool``

        :return:  list of node objects
        :rtype: ``list`` of :class: `Node`
        """
        data = self._paginated_request("/v2/instances", "instances")
        nodes = [self._to_node(item) for item in data]

        if ex_list_bare_metals:
            nodes += self.ex_list_bare_metal_nodes()
        return nodes

    def create_node(
        self,
        name: str,
        size: NodeSize,
        location: NodeLocation,
        image: Optional[NodeImage] = None,
        ex_ssh_key_ids: Optional[List[str]] = None,
        ex_private_network_ids: Optional[List[str]] = None,
        ex_snapshot: Union[VultrNodeSnapshot, str, None] = None,
        ex_enable_ipv6: bool = False,
        ex_backups: bool = False,
        ex_userdata: Optional[str] = None,
        ex_ddos_protection: bool = False,
        ex_enable_private_network: bool = False,
        ex_ipxe_chain_url: Optional[str] = None,
        ex_iso_id: Optional[str] = None,
        ex_script_id: Optional[str] = None,
        ex_image_id: Optional[str] = None,
        ex_activation_email: bool = False,
        ex_hostname: Optional[str] = None,
        ex_tag: Optional[str] = None,
        ex_firewall_group_id: Optional[str] = None,
        ex_reserved_ipv4: Optional[str] = None,
        ex_persistent_pxe: bool = False,
    ) -> Node:
        """Create a new node.

        :param name: The new node's name.
        :type name: ``str``

        :param size: The size to use to create the node.
        :type size: :class: `NodeSize`

        :param location: The location to provision the node.
        :type location: :class: `NodeLocation`

        :keyword image:  The image to use to provision the node.
        :type    image:  :class: `NodeImage`

        :keyword ex_ssh_key_ids: List of SSH keys to install on this node.
        :type    ex_ssh_key_ids: ``list`` of ``str``

        :keyword ex_private_network_ids: The network ids to attach to node.
                                         This parameter takes precedence over
                                         ex_enable_private_network (VPS only)
        :type    ex_private_network_ids: ``list`` of ``str``

        :keyword ex_snapshot: The snapshot to use when deploying the node.
                                 Mutually exclusive with image,
        :type    ex_snapshot: :class: `VultrNodeSnapshot` or ``str``

        :keyword ex_enable_ipv6: Whether to enable IPv6.
        :type    ex_enable_ipv6: ``bool``

        :keyword ex_backups: Enable automatic backups for the node. (VPS only)
        :type    ex_backups: ``bool``

        :keyword ex_userdata: String containing user data
        :type    ex_userdata: ``str``

        :keyword ex_ddos_protection: Enable DDoS protection (VPS only)
        :type    ex_ddos_protection: ``bool``

        :keyword ex_enable_private_network: Enable private networking.
                                            Mutually exclusive with
                                             ex_private_network_ids.
                                            (VPS only)
        :type    ex_enable_private_network: ``bool``

        :keyword ex_ipxe_chain_url: The URL location of the iPXE chainloader
                                    (VPS only)
        :type    ex_ipxe_chain_url: ``str``

        :keyword ex_iso_id: The ISO id to use when deploying this node.
                            (VPS only)
        :type    ex_iso_id: ``str``

        :keyword ex_script_id: The startup script id to use when deploying
                               this node.
        :type    ex_script_id: ``str``

        :keyword ex_image_id: The Application image_id to use when deploying
                              this node.
        :type    ex_image_id: ``str``

        :keyword ex_activation_email: Notify by email after deployment.
        :type    ex_activation_email: ``bool``

        :keyword ex_hostname: The hostname to use when deploying this node.
        :type    ex_hostname: ``str``

        :keyword ex_tag: The user-supplied tag.
        :type    ex_tag: ``str``

        :keyword ex_firewall_group_id: The Firewall Group id to attach to
                                       this node. (VPS only)
        :type    ex_firewall_group_id: ``str``

        :keyword ex_reserved_ipv4: Id of the floating IP to use as the
                                   main IP of this node.
        :type    ex_reserved_ipv4: ``str``

        :keyword ex_persistent_pxe: Enable persistent PXE (Bare Metal only)
        :type    ex_persistent_pxe: ``bool``
        """
        data = {
            "label": name,
            "region": location.id,
            "plan": size.id,
            "enable_ipv6": ex_enable_ipv6,
            "activation_email": ex_activation_email,
        }

        if image:
            data["os_id"] = image.id

        if ex_ssh_key_ids:
            data["sshkey_id"] = ex_ssh_key_ids

        if ex_snapshot:
            try:
                data["snapshot_id"] = ex_snapshot.id
            except AttributeError:
                data["snapshot_id"] = ex_snapshot

        if ex_userdata:
            data["user_data"] = base64.b64encode(bytes(ex_userdata, "utf-8")).decode("utf-8")

        if ex_script_id:
            data["script_id"] = ex_script_id

        if ex_image_id:
            data["image_id"] = ex_image_id

        if ex_hostname:
            data["hostname"] = ex_hostname

        if ex_reserved_ipv4:
            data["reserved_ipv4"] = ex_reserved_ipv4

        if ex_tag:
            data["tag"] = ex_tag

        if self._is_bare_metal(size):
            if ex_persistent_pxe:
                data["persistent_pxe"] = ex_persistent_pxe
            resp = self.connection.request("/v2/bare-metals", data=json.dumps(data), method="POST")
            return self._to_node(resp.object["bare_metal"])
        else:
            if ex_private_network_ids:
                data["attach_private_network"] = ex_private_network_ids

            if ex_enable_private_network:
                data["enable_private_network"] = ex_enable_private_network

            if ex_ipxe_chain_url:
                data["ipxe_chain_url"] = ex_ipxe_chain_url

            if ex_iso_id:
                data["iso_id"] = ex_iso_id

            if ex_ddos_protection:
                data["ddos_protection"] = ex_ddos_protection

            if ex_firewall_group_id:
                data["firewall_group_id"] = ex_firewall_group_id

            if ex_backups:
                data["backups"] = "enabled" if ex_backups is True else "disabled"

            resp = self.connection.request("/v2/instances", data=json.dumps(data), method="POST")
            return self._to_node(resp.object["instance"])

    def reboot_node(self, node: Node) -> bool:
        """Reboot the given node.

        :param node: The node to be rebooted.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        if self._is_bare_metal(node.size):
            return self.ex_reboot_bare_metal_node(node)

        resp = self.connection.request("/v2/instances/%s/reboot" % node.id, method="POST")

        return resp.success()

    def start_node(self, node: Node) -> bool:
        """Start the given node.

        :param node: The node to be started.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        if self._is_bare_metal(node.size):
            return self.ex_start_bare_metal_node(node)

        resp = self.connection.request("/v2/instances/%s/start" % node.id, method="POST")

        return resp.success()

    def stop_node(self, node: Node) -> bool:
        """Stop the given node.

        :param node: The node to be stopped.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        if self._is_bare_metal(node.size):
            return self.ex_stop_bare_metal_node(node)

        return self.ex_stop_nodes([node])

    def destroy_node(self, node: Node) -> bool:
        """Destroy the given node.

        :param node: The node to be destroyed.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        if self._is_bare_metal(node.size):
            return self.ex_destroy_bare_metal_node(node)

        resp = self.connection.request("/v2/instances/%s" % node.id, method="DELETE")

        return resp.success()

    def list_sizes(self, ex_list_bare_metals: bool = True) -> List[NodeSize]:
        """List available node sizes.

        :keyword ex_list_bare_metals: Whether to fetch bare metal sizes.
        :type    ex_list_bare_metals: ``bool``

        :rtype: ``list`` of :class: `NodeSize`
        """
        data = self._paginated_request("/v2/plans", "plans")
        sizes = [self._to_size(item) for item in data]

        if ex_list_bare_metals:
            sizes += self.ex_list_bare_metal_sizes()
        return sizes

    def list_images(self) -> List[NodeImage]:
        """List available node images.

        :rtype: ``list`` of :class: `NodeImage`
        """
        data = self._paginated_request("/v2/os", "os")
        return [self._to_image(item) for item in data]

    def list_locations(self) -> List[NodeLocation]:
        """List available node locations.

        :rtype: ``list`` of :class: `NodeLocation`
        """
        data = self._paginated_request("/v2/regions", "regions")
        return [self._to_location(item) for item in data]

    def list_volumes(self) -> List[StorageVolume]:
        """List storage volumes.

        :rtype: ``list`` of :class:`StorageVolume`
        """
        data = self._paginated_request("/v2/blocks", "blocks")
        return [self._to_volume(item) for item in data]

    def create_volume(
        self,
        size: int,
        name: str,
        location: Union[NodeLocation, str],
    ) -> StorageVolume:
        """Create a new volume.

        :param size: Size of the volume in gigabytes.\
        Size may range between 10 and 10000.
        :type size: ``int``

        :param name: Name of the volume to be created.
        :type name: ``str``

        :param location: Which data center to create the volume in.
        :type location: :class:`NodeLocation` or ``str``

        :return: The newly created volume.
        :rtype: :class:`StorageVolume`
        """

        data = {
            "label": name,
            "size_gb": size,
        }
        try:
            data["region"] = location.id
        except AttributeError:
            data["region"] = location

        resp = self.connection.request("/v2/blocks", data=json.dumps(data), method="POST")
        return self._to_volume(resp.object["block"])

    def attach_volume(
        self,
        node: Node,
        volume: StorageVolume,
        ex_live: bool = True,
    ) -> bool:
        """Attaches volume to node.

        :param node: Node to attach volume to.
        :type node: :class:`Node`

        :param volume: Volume to attach.
        :type volume: :class:`StorageVolume`

        :param ex_live: Attach the volume without restarting the node.
        :type ex_live: ``bool``

        :rytpe: ``bool``
        """

        data = {
            "instance_id": node.id,
            "live": ex_live,
        }
        resp = self.connection.request(
            "/v2/blocks/%s/attach" % volume.id, data=json.dumps(data), method="POST"
        )

        return resp.success()

    def detach_volume(
        self,
        volume: StorageVolume,
        ex_live: bool = True,
    ) -> bool:
        """Detaches a volume from a node.

        :param volume: Volume to be detached
        :type volume: :class:`StorageVolume`

        :param ex_live: Detach the volume without restarting the node.
        :type ex_live: ``bool``

        :rtype: ``bool``
        """
        data = {"live": ex_live}

        resp = self.connection.request(
            "/v2/blocks/%s/detach" % volume.id, data=json.dumps(data), method="POST"
        )

        return resp.success()

    def destroy_volume(self, volume: StorageVolume) -> bool:
        """Destroys a storage volume.

        :param volume: Volume to be destroyed
        :type  volume: :class:`StorageVolume`

        :rtype: ``bool``
        """

        resp = self.connection.request("/v2/blocks/%s" % volume.id, method="DELETE")

        return resp.success()

    def list_key_pairs(self) -> List[KeyPair]:
        """List all the available SSH key pair objects.

        :rtype: ``list`` of :class:`KeyPair`
        """
        data = self._paginated_request("/v2/ssh-keys", "ssh_keys")
        return [self._to_key_pair(item) for item in data]

    def get_key_pair(self, key_id: str) -> KeyPair:
        """Retrieve a single key pair.

        :param key_id: ID of the key pair to retrieve.
        :type  key_id: ``str``

        :rtype: :class: `KeyPair`
        """
        resp = self.connection.request("/v2/ssh-keys/%s" % key_id)
        return self._to_key_pair(resp.object["ssh_key"])

    def import_key_pair_from_string(self, name: str, key_material: str) -> KeyPair:
        """Import a new public key from string.

        :param name: Key pair name.
        :type  name: ``str``

        :param key_material: Public key material.
        :type  key_material: ``str``

        :rtype: :class: `KeyPair`
        """
        data = {
            "name": name,
            "ssh_key": key_material,
        }
        resp = self.connection.request("/v2/ssh-keys", data=json.dumps(data), method="POST")
        return self._to_key_pair(resp.object["ssh_key"])

    def delete_key_pair(self, key_pair: KeyPair) -> bool:
        """Delete existing key pair.

        :param key_pair: The key pair object to delete.
        :type key_pair: :class:`.KeyPair`

        :rtype: ``bool``
        """

        resp = self.connection.request("/v2/ssh-keys/%s" % key_pair.extra["id"], method="DELETE")

        return resp.success()

    def ex_list_bare_metal_nodes(self) -> List[Node]:
        """List all bare metal nodes.

        :return:  list of node objects
        :rtype: ``list`` of :class: `Node`
        """
        data = self._paginated_request("/v2/bare-metals", "bare_metals")
        return [self._to_node(item) for item in data]

    def ex_reboot_bare_metal_node(self, node: Node) -> bool:
        """Reboot the given bare metal node.

        :param node: The bare metal node to be rebooted.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        resp = self.connection.request("/v2/bare-metals/%s/reboot" % node.id, method="POST")

        return resp.success()

    def ex_resize_node(self, node: Node, size: NodeSize) -> bool:
        """Change size for the given node, only applicable for VPS nodes.

        :param node: The node to be resized.
        :type node: :class: `Node`

        :param size: The new size.
        :type size: :class: `NodeSize`
        """
        data = {"plan": size.id}
        resp = self.connection.request(
            "/v2/instances/%s" % node.id, data=json.dumps(data), method="PATCH"
        )
        return self._to_node(resp.object["instance"])

    def ex_start_bare_metal_node(self, node: Node) -> bool:
        """Start the given bare metal node.

        :param node: The bare metal node to be started.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        resp = self.connection.request("/v2/bare-metals/%s/start" % node.id, method="POST")

        return resp.success()

    def ex_stop_bare_metal_node(self, node: Node) -> bool:
        """Stop the given bare metal node.

        :param node: The bare metal node to be stopped.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        resp = self.connection.request("/v2/bare-metals/%s/halt" % node.id, method="POST")

        return resp.success()

    def ex_destroy_bare_metal_node(self, node: Node) -> bool:
        """Destroy the given bare metal node.

        :param node: The bare metal node to be destroyed.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        resp = self.connection.request("/v2/bare-metals/%s" % node.id, method="DELETE")

        return resp.success()

    def ex_get_node(self, node_id: str) -> Node:
        """Retrieve a node object.

        :param node_id: ID of the node to retrieve.
        :type snapshot_id: ``str``

        :rtype: :class: `Node`
        """
        resp = self.connection.request("/v2/instances/%s" % node_id)
        return self._to_node(resp.object["instance"])

    def ex_stop_nodes(self, nodes: List[Node]) -> bool:
        """Stops all the nodes given.

        : param nodes: A list of the nodes to stop.
        : type node: ``list`` of: class `Node`

        : rtype: ``bool``
        """

        data = {"instance_ids": [node.id for node in nodes]}
        resp = self.connection.request("/v2/instances/halt", data=json.dumps(data), method="POST")

        return resp.success()

    def ex_list_bare_metal_sizes(self) -> List[NodeSize]:
        """List bare metal sizes.

        :rtype: ``list`` of :class: `NodeSize`
        """
        data = self._paginated_request("/v2/plans-metal", "plans_metal")
        return [self._to_size(item) for item in data]

    def ex_list_snapshots(self) -> List[VultrNodeSnapshot]:
        """List node snapshots.

        :rtype: ``list`` of :class: `VultrNodeSnapshot`
        """
        data = self._paginated_request("/v2/snapshots", "snapshots")
        return [self._to_snapshot(item) for item in data]

    def ex_get_snapshot(self, snapshot_id: str) -> VultrNodeSnapshot:
        """Retrieve a snapshot.

        :param snapshot_id: ID of the snapshot to retrieve.
        :type snapshot_id: ``str``

        :rtype: :class: `VultrNodeSnapshot`
        """
        resp = self.connection.request("/v2/snapshots/%s" % snapshot_id)
        return self._to_snapshot(resp.object["snapshot"])

    def ex_create_snapshot(
        self, node: Node, description: Optional[str] = None
    ) -> VultrNodeSnapshot:
        """Create snapshot from a node.

        :param node: Node to create the snapshot from.
        :type node: :class: `Node`

        :keyword description: A description of the snapshot.
        :type    description: ``str``

        :rtype: :class: `VultrNodeSnapshot`
        """
        data = {
            "instance_id": node.id,
        }
        if description:
            data["description"] = description

        resp = self.connection.request("/v2/snapshots", data=json.dumps(data), method="POST")

        return self._to_snapshot(resp.object["snapshot"])

    def ex_delete_snapshot(self, snapshot: VultrNodeSnapshot) -> bool:
        """Delete the given snapshot.

        :param snapshot: The snapshot to delete.
        :type node: :class:`VultrNodeSnapshot`

        :rtype: ``bool``
        """

        resp = self.connection.request("/v2/snapshots/%s" % snapshot.id, method="DELETE")

        return resp.success()

    def ex_list_networks(self) -> List[VultrNetwork]:
        """List all private networks.

        :rtype: ``list`` of :class: `VultrNetwork`
        """

        data = self._paginated_request("/v2/private-networks", "networks")
        return [self._to_network(item) for item in data]

    def ex_create_network(
        self,
        cidr_block: str,
        location: Union[NodeLocation, str],
        description: Optional[str] = None,
    ) -> VultrNetwork:
        """Create a private network.

        :param cidr_block: The CIDR block assigned to the network.
        :type  cidr_block: ``str``

        :param location: The location to create the network.
        :type  location: :class: `NodeLocation` or ``str``

        :keyword description: A description of the private network.
        :type    description: ``str``

        :rtype: :class: `VultrNetwork`
        """
        subnet, subnet_mask = cidr_block.split("/")
        data = {
            "v4_subnet": subnet,
            "v4_subnet_mask": int(subnet_mask),
        }

        try:
            data["region"] = location.id
        except AttributeError:
            data["region"] = location

        if description:
            data["description"] = description

        resp = self.connection.request("/v2/private-networks", data=json.dumps(data), method="POST")

        return self._to_network(resp.object["network"])

    def ex_get_network(self, network_id: str) -> VultrNetwork:
        """Retrieve a private network.

        :param network_id: ID of the network to retrieve.
        :type network_id: ``str``

        :rtype: :class: `VultrNetwork`
        """

        resp = self.connection.request("/v2/private-networks/%s" % network_id)
        return self._to_network(resp.object["network"])

    def ex_destroy_network(self, network: VultrNetwork) -> bool:
        """Delete a private network.

        :param network: The network to destroy.
        :type  network: :class: `VultrNetwork`

        :rtype: ``bool``
        """
        resp = self.connection.request("/v2/private-networks/%s" % network.id, method="DELETE")

        return resp.success()

    def ex_list_available_sizes_for_location(
        self,
        location: NodeLocation,
    ) -> List[str]:
        """Get a list of available sizes for the given location.

        :param location: The location to get available sizes for.
        :type location: :class: `NodeLocation`

        :return:  A list of available size IDs for the given location.
        :rtype: ``list`` of ``str``
        """
        resp = self.connection.request("/v2/regions/%s/availability" % location.id)
        return resp.object["available_plans"]

    def ex_get_volume(self, volume_id: str) -> StorageVolume:
        """Retrieve a single volume.

        :param volume_id: The ID of the volume to fetch.
        :type volume_id: ``str``

        :rtype :class: `StorageVolume`
        :return: StorageVolume instance on success.
        """
        resp = self.connection.request("/v2/blocks/%s" % volume_id)

        return self._to_volume(resp.object["block"])

    def ex_resize_volume(self, volume: StorageVolume, size: int) -> bool:
        """Resize a volume.

        :param volume: The volume to resize.
        :type volume: :class:`StorageVolume`

        :param size: The new volume size in GBs.\
        Size may range between 10 and 10000.
        :type size: ``int``

        :rtype: ``bool``
        """
        data = {
            "label": volume.name,
            "size_gb": size,
        }

        resp = self.connection.request(
            "/v2/blocks/%s" % volume.id, data=json.dumps(data), method="PATCH"
        )
        return resp.success()

    def _is_bare_metal(self, size: Union[NodeSize, str]) -> bool:
        try:
            size_id = size.id
        except AttributeError:
            size_id = size

        return size_id.startswith("vbm")

    def _to_node(self, data: Dict[str, Any]) -> Node:
        id_ = data["id"]
        name = data["label"]
        public_ips = data["main_ip"].split() + data["v6_main_ip"].split()
        size = data["plan"]
        image = str(data["os_id"])
        created_at = data["date_created"]
        is_bare_metal = self._is_bare_metal(size)
        extra = {
            "location": data["region"],
            "ram": data["ram"],
            "disk": data["disk"],
            "netmask_v4": data["netmask_v4"],
            "gateway_v4": data["gateway_v4"],
            "v6_network": data["v6_network"],
            "v6_network_size": data["v6_network_size"],
            "app_id": data["app_id"],
            "image_id": data["image_id"],
            "features": data["features"],
            "tag": data["tag"],
            "os": data["os"],
            "is_bare_metal": is_bare_metal,
        }
        if is_bare_metal:
            state = self._get_node_state(data["status"])
            extra["cpu_count"] = data["cpu_count"]
            extra["mac_address"] = data["mac_address"]
            private_ips = None
        else:
            state = self._get_node_state(data["status"], power_state=data["power_status"])
            extra["vcpu_count"] = data["vcpu_count"]
            extra["allowed_bandwidth"] = data["allowed_bandwidth"]
            extra["power_status"] = data["power_status"]
            extra["server_status"] = data["server_status"]
            extra["firewall_group_id"] = data["firewall_group_id"]
            private_ips = data["internal_ip"].split()

        return Node(
            id=id_,
            name=name,
            state=state,
            public_ips=public_ips,
            private_ips=private_ips,
            driver=self,
            size=size,
            image=image,
            extra=extra,
            created_at=created_at,
        )

    def _to_volume(self, data: Dict[str, Any]) -> StorageVolume:
        id_ = data["id"]
        name = data["label"]
        size = data["size_gb"]
        try:
            state = self.VOLUME_STATE_MAP[data["status"]]
        except KeyError:
            state = StorageVolumeState.UNKNOWN
        extra = {
            "date_created": data["date_created"],
            "cost": data["cost"],
            "location": data["region"],
            "attached_to_instance": data["attached_to_instance"],
            "mount_id": data["mount_id"],
        }
        return StorageVolume(id=id_, name=name, size=size, driver=self, state=state, extra=extra)

    def _get_node_state(
        self,
        state: str,
        power_state: Optional[str] = None,
    ) -> NodeState:
        try:
            state = self.NODE_STATE_MAP[state]
        except KeyError:
            state = NodeState.UNKNOWN

        if power_state is None:
            return state

        if state == NodeState.RUNNING and power_state != "running":
            state = NodeState.STOPPED
        return state

    def _to_key_pair(self, data: Dict[str, Any]) -> KeyPair:
        name = data["name"]
        public_key = data["ssh_key"]
        # requires cryptography module
        try:
            fingerprint = get_pubkey_openssh_fingerprint(public_key)
        except RuntimeError:
            fingerprint = None
        extra = {
            "id": data["id"],
            "date_created": data["date_created"],
        }
        return KeyPair(
            name=name,
            public_key=public_key,
            fingerprint=fingerprint,
            driver=self,
            extra=extra,
        )

    def _to_location(self, data: Dict[str, Any]) -> NodeLocation:
        id_ = data["id"]
        name = data["city"]
        country = data["country"]
        extra = {
            "continent": data["continent"],
            "option": data["options"],
        }
        return NodeLocation(id=id_, name=name, country=country, driver=self, extra=extra)

    def _to_image(self, data: Dict[str, Any]) -> NodeImage:
        id_ = data["id"]
        name = data["name"]
        extra = {
            "arch": data["arch"],
            "family": data["family"],
        }
        return NodeImage(id=id_, name=name, driver=self, extra=extra)

    def _to_size(self, data: Dict[str, Any]) -> NodeSize:
        id_ = data["id"]
        ram = data["ram"]
        disk = data["disk"]
        bandwidth = data["bandwidth"]
        price = data["monthly_cost"]
        is_bare_metal = self._is_bare_metal(id_)
        extra = {
            "locations": data["locations"],
            "type": data["type"],
            "disk_count": data["disk_count"],
            "is_bare_metal": is_bare_metal,
        }

        # VPS and bare metal sizes have different fields
        if is_bare_metal is False:
            extra["vcpu_count"] = data["vcpu_count"]
        else:
            extra["cpu_count"] = data["cpu_count"]
            extra["cpu_model"] = data["cpu_model"]
            extra["cpu_threads"] = data["cpu_threads"]

        return NodeSize(
            id=id_,
            name=id_,
            ram=ram,
            disk=disk,
            bandwidth=bandwidth,
            price=price,
            driver=self,
            extra=extra,
        )

    def _to_network(self, data: Dict[str, Any]) -> VultrNetwork:
        id_ = data["id"]
        cidr_block = "{}/{}".format(data["v4_subnet"], data["v4_subnet_mask"])
        location = data["region"]
        extra = {
            "description": data["description"],
            "date_created": data["date_created"],
        }
        return VultrNetwork(id=id_, cidr_block=cidr_block, location=location, extra=extra)

    def _to_snapshot(self, data: Dict[str, Any]) -> VultrNodeSnapshot:
        id_ = data["id"]
        created = data["date_created"]
        # Size is returned in bytes, convert to GBs
        size = data["size"] / 1024 / 1024 / 1024
        try:
            state = self.SNAPSHOT_STATE_MAP[data["status"]]
        except KeyError:
            state = VolumeSnapshotState.UNKNOWN
        extra = {
            "description": data["description"],
            "os_id": data["os_id"],
            "app_id": data["app_id"],
        }
        return VultrNodeSnapshot(
            id=id_, size=size, created=created, state=state, extra=extra, driver=self
        )

    def _paginated_request(
        self, url: str, key: str, params: Optional[Dict[str, Any]] = None
    ) -> List[Any]:
        """Perform multiple calls to get the full list of items when
        the API responses are paginated.

        :param url: API endpoint
        :type url: ``str``

        :param key: Result object key
        :type key: ``str``

        :param params: Request parameters
        :type params: ``dict``

        :return: ``list`` of API response objects
        :rtype: ``list``
        """
        params = params if params is not None else {}
        resp = self.connection.request(url, params=params).object
        data = list(resp.get(key, []))
        objects = data
        while True:
            next_page = resp["meta"]["links"]["next"]
            if next_page:
                params["cursor"] = next_page
                resp = self.connection.request(url, params=params).object
                data = list(resp.get(key, []))
                objects.extend(data)
            else:
                return objects
